# Standard library imports.
import unittest

# Library imports.
import pytest
import numpy as np

# Local imports.
from pysph.base.utils import get_particle_array
from compyle.config import get_config
from compyle.api import declare
from compyle.profile import get_profile_info

from pysph.sph.equation import Equation, Group
from pysph.sph.acceleration_eval import (
    AccelerationEval, MegaGroup, CythonGroup,
    check_equation_array_properties
)
from pysph.sph.basic_equations import SummationDensity
from pysph.base.kernels import CubicSpline
from pysph.base.nnps import LinkedListNNPS as NNPS
from pysph.sph.sph_compiler import SPHCompiler

from pysph.base.reduce_array import serial_reduce_array


class DummyEquation(Equation):
    def initialize(self, d_idx, d_rho, d_V):
        d_rho[d_idx] = d_V[d_idx]

    def loop(self, d_idx, d_rho, s_idx, s_m, s_u, WIJ):
        d_rho[d_idx] += s_m[s_idx] * WIJ

    def post_loop(self, d_idx, d_rho, s_idx, s_m, s_V):
        d_rho[d_idx] += s_m[d_idx]


class FindTotalMass(Equation):
    def initialize(self, d_idx, d_m, d_total_mass):
        # FIXME: This is stupid and should be fixed if we add a separate
        # initialize_once function or so.
        d_total_mass[0] = 0.0

    def post_loop(self, d_idx, d_m, d_total_mass):
        d_total_mass[0] += d_m[d_idx]


class TestCheckEquationArrayProps(unittest.TestCase):
    def test_should_raise_runtime_error_when_invalid_dest_source(self):
        # Given
        f = get_particle_array(name='f')

        # When
        eq = SummationDensity(dest='fluid', sources=['f'])

        # Then
        self.assertRaises(
            RuntimeError,
            check_equation_array_properties,
            eq, [f]
        )

        # When
        eq = SummationDensity(dest='f', sources=['fluid'])

        # Then
        self.assertRaises(
            RuntimeError,
            check_equation_array_properties,
            eq, [f]
        )

    def test_should_pass_when_properties_exist(self):
        # Given
        f = get_particle_array(name='f')

        # When
        eq = SummationDensity(dest='f', sources=['f'])

        # Then
        check_equation_array_properties(eq, [f])

    def test_should_fail_when_props_dont_exist(self):
        # Given
        f = get_particle_array(name='f')

        # When
        eq = DummyEquation(dest='f', sources=['f'])

        # Then
        self.assertRaises(RuntimeError,
                          check_equation_array_properties, eq, [f])

    def test_should_fail_when_src_props_dont_exist(self):
        # Given
        f = get_particle_array(name='f')
        f.add_property('V')
        s = get_particle_array(name='s')

        # When
        eq = DummyEquation(dest='f', sources=['f', 's'])

        # Then
        self.assertRaises(RuntimeError,
                          check_equation_array_properties, eq, [f, s])

    def test_should_pass_when_src_props_exist(self):
        # Given
        f = get_particle_array(name='f')
        f.add_property('V')
        s = get_particle_array(name='s')
        s.add_property('V')

        # When
        eq = DummyEquation(dest='f', sources=['f', 's'])

        # Then
        check_equation_array_properties(eq, [f, s])

    def test_should_check_constants(self):
        # Given
        f = get_particle_array(name='f')

        # When
        eq = FindTotalMass(dest='f', sources=['f'])

        # Then.
        self.assertRaises(RuntimeError,
                          check_equation_array_properties, eq, [f])

        # When.
        f.add_constant('total_mass', 0.0)

        # Then.
        check_equation_array_properties(eq, [f])


class SimpleEquation(Equation):
    def __init__(self, dest, sources):
        super(SimpleEquation, self).__init__(dest, sources)
        self.count = 0

    def initialize(self, d_idx, d_u, d_au):
        d_u[d_idx] = 0.0
        d_au[d_idx] = 0.0

    def loop(self, d_idx, d_au, s_idx, s_m):
        d_au[d_idx] += s_m[s_idx]

    def post_loop(self, d_idx, d_u, d_au):
        d_u[d_idx] = d_au[d_idx]

    def converged(self):
        self.count += 1
        result = self.count - 1
        if result > 0:
            # Reset the count for the next loop.
            self.count = 0
        return result


class MixedTypeEquation(Equation):
    def initialize(self, d_idx, d_u, d_au, d_pid, d_tag):
        d_u[d_idx] = 0.0 + d_pid[d_idx]
        d_au[d_idx] = 0.0 + d_tag[d_idx]

    def loop(self, d_idx, d_au, s_idx, s_m, s_pid, s_tag):
        d_au[d_idx] += s_m[s_idx] + s_pid[s_idx] + s_tag[s_idx]

    def post_loop(self, d_idx, d_u, d_au, d_pid):
        d_u[d_idx] = d_au[d_idx] + d_pid[d_idx]


class SimpleReduction(Equation):
    def initialize(self, d_idx, d_au):
        d_au[d_idx] = 0.0

    def reduce(self, dst, t, dt):
        dst.total_mass[0] = serial_reduce_array(dst.m, op='sum')
        if dst.gpu is not None:
            dst.gpu.push('total_mass')


class PyInit(Equation):
    def py_initialize(self, dst, t, dt):
        self.called_with = t, dt
        if dst.gpu:
            dst.gpu.pull('au')
        dst.au[:] = 1.0
        if dst.gpu:
            dst.gpu.push('au')

    def initialize(self, d_idx, d_au):
        d_au[d_idx] += 1.0


class LoopAllEquation(Equation):
    def initialize(self, d_idx, d_rho):
        d_rho[d_idx] = 0.0

    def loop(self, d_idx, d_rho, s_m, s_idx, WIJ):
        d_rho[d_idx] += s_m[s_idx] * WIJ

    def loop_all(self, d_idx, d_x, d_rho, s_m, s_x, s_h, SPH_KERNEL, NBRS,
                 N_NBRS):
        i = declare('int')
        s_idx = declare('long')
        xij = declare('matrix((3,))')
        rij = 0.0
        sum = 0.0
        xij[1] = 0.0
        xij[2] = 0.0
        for i in range(N_NBRS):
            s_idx = NBRS[i]
            xij[0] = d_x[d_idx] - s_x[s_idx]
            rij = abs(xij[0])
            sum += s_m[s_idx] * SPH_KERNEL.kernel(xij, rij, s_h[s_idx])
        d_rho[d_idx] += sum


class DumbEquation(Equation):
    def initialize(self, d_idx, d_au):
        d_au[d_idx] += 1

    def loop(self, d_idx, d_au):
        d_au[d_idx] += 1

    def post_loop(self, d_idx, d_au):
        d_au[d_idx] += 1

    def reduce(self, dst, t, dt):
        dst.reduce_calls[0] = dst.reduce_calls[0] + 1


class InitializePair(Equation):
    def initialize_pair(self, d_idx, d_u, s_u):
        # Will only work if the source/destinations are the same
        # but should do for a test.
        d_u[d_idx] = s_u[d_idx]*1.5


class TestMegaGroup(unittest.TestCase):
    def test_ensure_group_retains_user_order_of_equations(self):
        # Given
        group = Group(equations=[
            SimpleEquation(dest='f', sources=['s', 'f']),
            DummyEquation(dest='f', sources=['s', 'f']),
            MixedTypeEquation(dest='f', sources=['f']),
        ])

        # When
        mg = MegaGroup(group, CythonGroup)

        # Then
        data = mg.data
        self.assertEqual(list(data.keys()), ['f'])
        eqs_with_no_source, sources, all_eqs = data['f']
        self.assertEqual(len(eqs_with_no_source.equations), 0)

        all_eqs_order = [x.__class__.__name__ for x in all_eqs.equations]
        expect = ['SimpleEquation', 'DummyEquation', 'MixedTypeEquation']
        self.assertEqual(all_eqs_order, expect)

        self.assertEqual(sorted(sources.keys()), ['f', 's'])
        s_eqs = [x.__class__.__name__ for x in sources['s'].equations]
        expect = ['SimpleEquation', 'DummyEquation']
        self.assertEqual(s_eqs, expect)

        f_eqs = [x.__class__.__name__ for x in sources['f'].equations]
        expect = ['SimpleEquation', 'DummyEquation', 'MixedTypeEquation']
        self.assertEqual(f_eqs, expect)

    def test_mega_group_copies_props_of_group(self):
        # Given
        def nothing():
            pass

        g = Group(
            equations=[], real=False, update_nnps=True, iterate=True,
            max_iterations=20, min_iterations=2, pre=nothing, post=nothing,
            start_idx=1, stop_idx=2
        )

        # When
        mg = MegaGroup(g, CythonGroup)

        # Then
        props = ('real update_nnps iterate max_iterations condition '
                 'min_iterations pre post start_idx stop_idx').split()
        for prop in props:
            self.assertEqual(getattr(mg, prop), getattr(g, prop))


class TestAccelerationEval1D(unittest.TestCase):
    def setUp(self):
        self.dim = 1
        n = 10
        dx = 1.0 / (n - 1)
        x = np.linspace(0, 1, n)
        m = np.ones_like(x)
        h = np.ones_like(x) * dx * 1.05
        pa = get_particle_array(name='fluid', x=x, h=h, m=m)
        self.pa = pa

    def _make_accel_eval(self, equations, cache_nnps=False):
        arrays = [self.pa]
        kernel = CubicSpline(dim=self.dim)
        a_eval = AccelerationEval(
            particle_arrays=arrays, equations=equations, kernel=kernel
        )
        comp = SPHCompiler(a_eval, integrator=None)
        comp.compile()
        nnps = NNPS(dim=kernel.dim, particles=arrays, cache=cache_nnps)
        nnps.update()
        a_eval.set_nnps(nnps)
        return a_eval

    def test_should_support_constants(self):
        # Given
        pa = self.pa
        pa.add_constant('total_mass', 0.0)
        equations = [FindTotalMass(dest='fluid', sources=['fluid'])]
        a_eval = self._make_accel_eval(equations)

        # When
        a_eval.compute(0.1, 0.1)

        # Then
        self.assertEqual(pa.total_mass, 10.0)

    def test_should_not_iterate_normal_group(self):
        # Given
        pa = self.pa
        equations = [SimpleEquation(dest='fluid', sources=['fluid'])]
        a_eval = self._make_accel_eval(equations)

        # When
        a_eval.compute(0.1, 0.1)

        # Then
        expect = np.asarray([3., 4., 5., 5., 5., 5., 5., 5., 4., 3.])
        self.assertListEqual(list(pa.u), list(expect))

    def test_should_work_with_cached_nnps(self):
        # Given
        pa = self.pa
        equations = [SimpleEquation(dest='fluid', sources=['fluid'])]
        a_eval = self._make_accel_eval(equations, cache_nnps=True)

        # When
        a_eval.compute(0.1, 0.1)

        # Then
        expect = np.asarray([3., 4., 5., 5., 5., 5., 5., 5., 4., 3.])
        self.assertListEqual(list(pa.u), list(expect))

    def test_should_iterate_iterated_group(self):
        # Given
        pa = self.pa
        equations = [Group(
            equations=[
                SimpleEquation(dest='fluid', sources=['fluid']),
                SimpleEquation(dest='fluid', sources=['fluid']),
            ],
            iterate=True
        )]
        a_eval = self._make_accel_eval(equations)

        # When
        a_eval.compute(0.1, 0.1)

        # Then
        expect = np.asarray([3., 4., 5., 5., 5., 5., 5., 5., 4., 3.]) * 2
        self.assertListEqual(list(pa.u), list(expect))

    def test_should_iterate_nested_groups(self):
        pa = self.pa
        equations = [Group(
            equations=[
                Group(
                    equations=[SimpleEquation(dest='fluid', sources=['fluid'])]
                ),
                Group(
                    equations=[SimpleEquation(dest='fluid', sources=['fluid'])]
                ),
            ],
            iterate=True,
        )]
        a_eval = self._make_accel_eval(equations)

        # When
        a_eval.compute(0.1, 0.1)

        # Then
        expect = np.asarray([3., 4., 5., 5., 5., 5., 5., 5., 4., 3.])
        self.assertListEqual(list(pa.u), list(expect))

    def test_should_run_reduce(self):
        # Given.
        pa = self.pa
        pa.add_constant('total_mass', 0.0)
        equations = [SimpleReduction(dest='fluid', sources=['fluid'])]
        a_eval = self._make_accel_eval(equations)

        # When
        a_eval.compute(0.1, 0.1)

        # Then
        expect = np.sum(pa.m)
        self.assertAlmostEqual(pa.total_mass[0], expect, 14)

    def test_should_call_initialize_pair(self):
        # Given.
        pa = self.pa
        pa.u[:] = 1.0
        if pa.gpu:
            pa.gpu.push('u')
        equations = [InitializePair(dest='fluid', sources=['fluid'])]
        a_eval = self._make_accel_eval(equations)

        # When
        a_eval.compute(0.0, 0.1)

        # Then
        if pa.gpu:
            pa.gpu.pull('u')
        np.testing.assert_array_almost_equal(
            pa.u, np.ones_like(pa.x)*1.5
        )

    def test_should_call_py_initialize(self):
        # Given.
        pa = self.pa
        equations = [PyInit(dest='fluid', sources=None)]
        eq = equations[0]
        a_eval = self._make_accel_eval(equations)

        # When
        a_eval.compute(1.0, 0.1)

        # Then
        if pa.gpu:
            pa.gpu.pull('au')
        np.testing.assert_array_almost_equal(
            pa.au, np.ones_like(pa.x) * 2.0
        )
        self.assertEqual(eq.called_with, (1.0, 0.1))

    def test_should_work_with_non_double_arrays(self):
        # Given
        pa = self.pa
        equations = [MixedTypeEquation(dest='fluid', sources=['fluid'])]
        a_eval = self._make_accel_eval(equations)

        # When
        a_eval.compute(0.1, 0.1)

        # Then
        expect = np.asarray([3., 4., 5., 5., 5., 5., 5., 5., 4., 3.])
        self.assertListEqual(list(pa.u), list(expect))

    def test_should_support_loop_all_and_loop(self):
        # Given
        pa = self.pa
        equations = [SummationDensity(dest='fluid', sources=['fluid'])]
        a_eval = self._make_accel_eval(equations)
        a_eval.compute(0.1, 0.1)
        ref_rho = pa.rho.copy()

        # When
        pa.rho[:] = 0.0
        equations = [LoopAllEquation(dest='fluid', sources=['fluid'])]
        a_eval = self._make_accel_eval(equations)
        a_eval.compute(0.1, 0.1)

        # Then
        # 2*ref_rho as we are doing both the loop and loop_all to test if
        # both are called.
        self.assertTrue(np.allclose(pa.rho, 2.0 * ref_rho))

    def test_should_handle_repeated_helper_functions(self):
        pa = self.pa

        def helper(x=1.0):
            return x * 1.5

        class SillyEquation2(Equation):
            def initialize(self, d_idx, d_au, d_m):
                d_au[d_idx] += helper(d_m[d_idx])

            def _get_helpers_(self):
                return [helper]

        equations = [SillyEquation2(dest='fluid', sources=['fluid']),
                     SillyEquation2(dest='fluid', sources=['fluid'])]

        a_eval = self._make_accel_eval(equations)

        # When
        a_eval.compute(0.1, 0.1)

        # Then
        expect = np.ones(10) * 3.0
        self.assertListEqual(list(pa.au), list(expect))

    def test_should_call_pre_post_functions_in_group(self):
        # Given
        pa = self.pa

        def pre():
            pa.m += 1.0

        def post():
            pa.u += 1.0

        equations = [
            Group(
                equations=[
                    SimpleEquation(dest='fluid', sources=['fluid'])
                ],
                pre=pre, post=post
            )
        ]
        a_eval = self._make_accel_eval(equations)

        # When
        a_eval.compute(0.1, 0.1)

        # Then
        expect = np.asarray([7., 9., 11., 11., 11., 11., 11., 11., 9., 7.])
        self.assertListEqual(list(pa.u), list(expect))

    def test_should_honor_start_stop_idx_in_group(self):
        # Given
        pa = self.pa
        pa.u[:] = 1.0
        pa.au[:] = 1.0

        equations = [
            Group(
                equations=[
                    SimpleEquation(dest='fluid', sources=['fluid'])
                ],
                start_idx=1, stop_idx=2
            )
        ]
        a_eval = self._make_accel_eval(equations)

        # When
        a_eval.compute(0.1, 0.1)

        # Then
        expect = np.ones_like(pa.u)
        expect[1] = 4.0
        self.assertListEqual(list(pa.u), list(expect))
        self.assertListEqual(list(pa.au), list(expect))

    def test_should_honor_start_stop_idx_as_str_in_group(self):
        # Given
        pa = self.pa
        pa.add_constant('start', 1)
        pa.add_constant('stop', 3)
        pa.u[:] = 1.0
        pa.au[:] = 1.0

        equations = [
            Group(
                equations=[
                    SimpleEquation(dest='fluid', sources=['fluid'])
                ],
                start_idx='start', stop_idx='stop'
            )
        ]
        a_eval = self._make_accel_eval(equations)

        # When
        a_eval.compute(0.1, 0.1)

        # Then
        expect = np.ones_like(pa.u)
        expect[1] = 4.0
        expect[2] = 5.0
        self.assertListEqual(list(pa.u), list(expect))
        self.assertListEqual(list(pa.au), list(expect))

    def test_group_honors_condition(self):
        # Given
        pa = self.pa
        pa.add_constant('reduce_calls', 0)
        pa.au[:] = 0.0
        if pa.gpu:
            pa.gpu.push('au')
        call_data = []

        def cond(t, dt):
            call_data.append((t, dt))
            return False

        equations = [
            Group(
                equations=[
                    DumbEquation(dest='fluid', sources=['fluid'])
                ],
                condition=cond
            ),
            Group(
                equations=[
                    Group(
                        equations=[
                            DumbEquation(dest='fluid', sources=['fluid'])
                        ],
                        condition=cond
                    )
                ]
            ),
            Group(
                equations=[
                    DumbEquation(dest='fluid', sources=['fluid'])
                ],
            )
        ]
        a_eval = self._make_accel_eval(equations)

        # When
        a_eval.compute(0.0, 0.1)

        # Then
        if pa.gpu:
            pa.gpu.pull('au')

        expect = np.ones_like(pa.au)*7
        expect[0] = expect[-1] = 5
        expect[1] = expect[-2] = 6
        self.assertEqual(len(call_data), 2)
        self.assertEqual(call_data[0], (0.0, 0.1))
        self.assertEqual(call_data[1], (0.0, 0.1))
        self.assertListEqual(list(pa.au), list(expect))


class EqWithTime(Equation):
    def initialize(self, d_idx, d_au, t, dt):
        d_au[d_idx] = t + dt

    def loop(self, d_idx, d_au, s_idx, s_m, t, dt):
        d_au[d_idx] += t + dt


class TestAccelerationEval1DGPU(unittest.TestCase):
    # Fix this to be a subclass of TestAccelerationEval1D

    def setUp(self):
        self.dim = 1
        n = 10
        dx = 1.0 / (n - 1)
        x = np.linspace(0, 1, n)
        m = np.ones_like(x)
        h = np.ones_like(x) * dx * 1.05
        pa = get_particle_array(name='fluid', x=x, h=h, m=m)
        self.pa = pa

    def _get_nnps_cls(self):
        from pysph.base.gpu_nnps import ZOrderGPUNNPS as GPUNNPS
        return GPUNNPS

    def _make_accel_eval(self, equations, cache_nnps=True):
        pytest.importorskip('pyopencl')
        pytest.importorskip('pysph.base.gpu_nnps')
        GPUNNPS = self._get_nnps_cls()
        arrays = [self.pa]
        kernel = CubicSpline(dim=self.dim)
        a_eval = AccelerationEval(
            particle_arrays=arrays, equations=equations, kernel=kernel,
            backend='opencl'
        )
        comp = SPHCompiler(a_eval, integrator=None)
        comp.compile()
        self.sph_compiler = comp
        nnps = GPUNNPS(dim=kernel.dim, particles=arrays, cache=cache_nnps,
                       backend='opencl')
        nnps.update()
        a_eval.set_nnps(nnps)
        return a_eval

    def test_accel_eval_should_work_on_gpu(self):
        # Given
        pa = self.pa
        equations = [SimpleEquation(dest='fluid', sources=['fluid'])]
        a_eval = self._make_accel_eval(equations)

        # When
        a_eval.compute(0.1, 0.1)

        # Then
        expect = np.asarray([3., 4., 5., 5., 5., 5., 5., 5., 4., 3.])
        pa.gpu.pull('u')
        self.assertListEqual(list(pa.u), list(expect))

    def test_precomputed_should_work_on_gpu(self):
        # Given
        pa = self.pa
        equations = [SummationDensity(dest='fluid', sources=['fluid'])]
        a_eval = self._make_accel_eval(equations)

        # When
        a_eval.compute(0.1, 0.1)

        # Then
        expect = np.asarray([7.357, 9.0, 9., 9., 9., 9., 9., 9., 9., 7.357])
        pa.gpu.pull('rho')

        print(pa.rho, pa.gpu.rho)
        self.assertTrue(np.allclose(expect, pa.rho, atol=1e-2))

    def test_precomputed_should_work_on_gpu_with_double(self):
        orig = get_config().use_double

        def _cleanup():
            get_config().use_double = orig

        get_config().use_double = True
        self.addCleanup(_cleanup)
        # Given
        pa = self.pa
        equations = [SummationDensity(dest='fluid', sources=['fluid'])]
        a_eval = self._make_accel_eval(equations)

        # When
        a_eval.compute(0.1, 0.1)

        # Then
        expect = np.asarray([7.357, 9.0, 9., 9., 9., 9., 9., 9., 9., 7.357])
        pa.gpu.pull('rho')

        print(pa.rho, pa.gpu.rho)
        self.assertTrue(np.allclose(expect, pa.rho, atol=1e-2))

    def test_equation_with_time_should_work_on_gpu(self):
        # Given
        pa = self.pa
        equations = [EqWithTime(dest='fluid', sources=['fluid'])]
        a_eval = self._make_accel_eval(equations)

        # When
        a_eval.compute(0.2, 0.1)

        # Then
        expect = np.asarray([4., 5., 6., 6., 6., 6., 6., 6., 5., 4.]) * 0.3
        pa.gpu.pull('au')
        print(pa.au, expect)
        self.assertTrue(np.allclose(expect, pa.au))

    def test_update_nnps_is_called_on_gpu(self):
        # Given
        equations = [
            Group(
                equations=[
                    SummationDensity(dest='fluid', sources=['fluid']),
                ],
                update_nnps=True
            ),
            Group(
                equations=[EqWithTime(dest='fluid', sources=['fluid'])]
            ),
        ]

        # When
        a_eval = self._make_accel_eval(equations)

        # Then
        h = a_eval.c_acceleration_eval.helper
        assert len(h.calls) == 5
        call = h.calls[0]
        assert call['type'] == 'kernel'
        assert call['method'].function_name == 'g0_fluid_initialize'
        assert call['loop'] is False

        call = h.calls[1]
        assert call['type'] == 'kernel'
        assert call['method'].function_name == 'g0_fluid_on_fluid_loop'
        assert call['loop'] is True

        call = h.calls[2]
        assert call['type'] == 'method'
        assert call['method'] == 'update_nnps'

        call = h.calls[3]
        assert call['type'] == 'kernel'
        assert call['method'].function_name == 'g1_fluid_initialize'
        assert call['loop'] is False

        call = h.calls[4]
        assert call['type'] == 'kernel'
        assert call['method'].function_name == 'g1_fluid_on_fluid_loop'
        assert call['loop'] is True

    def test_should_stop_iteration_with_max_iteration_on_gpu(self):
        pa = self.pa

        class SillyEquation(Equation):
            def loop(self, d_idx, d_au, s_idx, s_m):
                d_au[d_idx] += s_m[s_idx]

            def converged(self):
                return 0

        equations = [Group(
            equations=[
                Group(
                    equations=[SillyEquation(dest='fluid', sources=['fluid'])]
                ),
                Group(
                    equations=[SillyEquation(dest='fluid', sources=['fluid'])]
                ),
            ],
            iterate=True, max_iterations=2,
        )]
        a_eval = self._make_accel_eval(equations)

        # When
        a_eval.compute(0.1, 0.1)

        # Then
        expect = np.asarray([3., 4., 5., 5., 5., 5., 5., 5., 4., 3.]) * 4.0
        pa.gpu.pull('au')
        self.assertListEqual(list(pa.au), list(expect))

    def test_should_stop_iteration_with_converged_on_gpu(self):
        pa = self.pa

        class SillyEquation1(Equation):
            def __init__(self, dest, sources):
                super(SillyEquation1, self).__init__(dest, sources)
                self.conv = 0

            def loop(self, d_idx, d_au, s_idx, s_m):
                d_au[d_idx] += s_m[s_idx]

            def post_loop(self, d_idx, d_au):
                if d_au[d_idx] > 19.0:
                    self.conv = 1

            def converged(self):
                if hasattr(self, '_pull'):
                    # _pull is not available on CPU.
                    self._pull('conv')
                return self.conv

        equations = [Group(
            equations=[
                Group(
                    equations=[SillyEquation1(dest='fluid', sources=['fluid'])]
                ),
                Group(
                    equations=[SillyEquation1(dest='fluid', sources=['fluid'])]
                ),
            ],
            iterate=True, max_iterations=10,
        )]
        a_eval = self._make_accel_eval(equations)

        # When
        a_eval.compute(0.1, 0.1)

        # Then
        expect = np.asarray([3., 4., 5., 5., 5., 5., 5., 5., 4., 3.]) * 6.0
        pa.gpu.pull('au')
        self.assertListEqual(list(pa.au), list(expect))

    def test_should_handle_helper_functions_on_gpu(self):
        pa = self.pa

        def helper(x=1.0):
            return x * 1.5

        class SillyEquation2(Equation):
            def initialize(self, d_idx, d_au, d_m):
                d_au[d_idx] += helper(d_m[d_idx])

            def _get_helpers_(self):
                return [helper]

        equations = [SillyEquation2(dest='fluid', sources=['fluid']),
                     SillyEquation2(dest='fluid', sources=['fluid'])]

        a_eval = self._make_accel_eval(equations)

        # When
        a_eval.compute(0.1, 0.1)

        # Then
        expect = np.ones(10) * 3.0
        pa.gpu.pull('au')
        self.assertListEqual(list(pa.au), list(expect))

    def test_should_run_reduce_when_using_gpu(self):
        # Given.
        pa = self.pa
        pa.add_constant('total_mass', 0.0)
        equations = [SimpleReduction(dest='fluid', sources=['fluid'])]
        a_eval = self._make_accel_eval(equations)

        # When
        a_eval.compute(0.1, 0.1)

        # Then
        expect = np.sum(pa.m)
        pa.gpu.pull('total_mass')
        self.assertAlmostEqual(pa.total_mass[0], expect, 14)

    def test_should_call_initialize_pair_on_gpu(self):
        # Given.
        pa = self.pa
        pa.u[:] = 1.0
        if pa.gpu:
            pa.gpu.push('u')
        equations = [InitializePair(dest='fluid', sources=['fluid'])]
        a_eval = self._make_accel_eval(equations)

        # When
        a_eval.compute(0.0, 0.1)

        # Then
        if pa.gpu:
            pa.gpu.pull('u')
        np.testing.assert_array_almost_equal(
            pa.u, np.ones_like(pa.x)*1.5
        )

    def test_should_call_py_initialize_for_gpu_backend(self):
        # Given.
        pa = self.pa
        equations = [PyInit(dest='fluid', sources=None)]
        eq = equations[0]
        a_eval = self._make_accel_eval(equations)

        # When
        a_eval.compute(1.0, 0.1)

        # Then
        if pa.gpu:
            pa.gpu.pull('au')
        np.testing.assert_array_almost_equal(
            pa.au, np.ones_like(pa.x) * 2.0
        )
        self.assertEqual(eq.called_with, (1.0, 0.1))

    def test_get_equations_with_converged(self):
        pytest.importorskip('pyopencl')
        pytest.importorskip('pysph.base.gpu_nnps')
        from pysph.sph.acceleration_eval_gpu_helper import \
            get_equations_with_converged
        # Given
        se = SimpleEquation(dest='fluid', sources=['fluid'])
        se1 = SimpleEquation(dest='fluid', sources=['fluid'])
        sd = SummationDensity(dest='fluid', sources=['fluid'])
        me = MixedTypeEquation(dest='fluid', sources=['fluid'])
        eq_t = EqWithTime(dest='fluid', sources=['fluid'])
        g = Group(
            equations=[
                Group(equations=[Group(equations=[se, sd])],
                      iterate=True, max_iterations=10),
                Group(equations=[me, eq_t, se1]),
            ],
        )

        # When
        eqs = get_equations_with_converged(g)

        # Then
        assert eqs == [se, se1]

    def test_should_support_loop_all_and_loop_on_gpu(self):
        # Given
        pa = self.pa
        equations = [SummationDensity(dest='fluid', sources=['fluid'])]
        a_eval = self._make_accel_eval(equations)
        a_eval.compute(0.1, 0.1)
        pa.gpu.pull('rho')
        ref_rho = pa.rho.copy()

        # When
        pa.rho[:] = 0.0
        pa.gpu.push('rho')
        equations = [LoopAllEquation(dest='fluid', sources=['fluid'])]
        a_eval = self._make_accel_eval(equations)
        a_eval.compute(0.1, 0.1)

        # Then
        # 2*ref_rho as we are doing both the loop and loop_all to test if
        # both are called.
        pa.gpu.pull('rho')

        self.assertTrue(np.allclose(pa.rho, 2.0 * ref_rho))

    def test_should_call_pre_post_functions_in_group_on_gpu(self):
        # Given
        pa = self.pa

        def pre():
            pa.m += 1.0
            pa.gpu.push('m')

        def post():
            pa.gpu.pull('u')
            pa.u += 1.0

        equations = [
            Group(
                equations=[
                    SimpleEquation(dest='fluid', sources=['fluid'])
                ],
                pre=pre, post=post

            )
        ]
        a_eval = self._make_accel_eval(equations)

        # When
        a_eval.compute(0.1, 0.1)

        # Then
        expect = np.asarray([7., 9., 11., 11., 11., 11., 11., 11., 9., 7.])
        self.assertListEqual(list(pa.u), list(expect))

    def test_should_honor_start_stop_idx_in_group_on_gpu(self):
        # Given
        pa = self.pa
        pa.u[:] = 1.0
        pa.au[:] = 1.0
        if pa.gpu:
            pa.gpu.push('u', 'au')

        equations = [
            Group(
                equations=[
                    SimpleEquation(dest='fluid', sources=['fluid'])
                ],
                start_idx=1, stop_idx=2
            )
        ]
        a_eval = self._make_accel_eval(equations)

        # When
        a_eval.compute(0.1, 0.1)

        # Then
        if pa.gpu:
            pa.gpu.pull('u', 'au')
        expect = np.ones_like(pa.u)
        expect[1] = 4.0
        self.assertListEqual(list(pa.u), list(expect))
        self.assertListEqual(list(pa.au), list(expect))

    def test_should_honor_start_stop_idx_as_str_in_group(self):
        # Given
        pa = self.pa
        pa.add_constant('start', 1)
        pa.add_constant('stop', 3)
        pa.u[:] = 1.0
        pa.au[:] = 1.0
        if pa.gpu:
            pa.gpu.push('u', 'au')

        equations = [
            Group(
                equations=[
                    SimpleEquation(dest='fluid', sources=['fluid'])
                ],
                start_idx='start', stop_idx='stop'
            )
        ]
        a_eval = self._make_accel_eval(equations)

        # When
        a_eval.compute(0.1, 0.1)

        # Then
        if pa.gpu:
            pa.gpu.pull('u', 'au')

        expect = np.ones_like(pa.u)
        expect[1] = 4.0
        expect[2] = 5.0
        self.assertListEqual(list(pa.u), list(expect))
        self.assertListEqual(list(pa.au), list(expect))

    def test_group_honors_condition_on_gpu(self):
        # Given
        pa = self.pa
        pa.add_constant('reduce_calls', 0)
        pa.au[:] = 0.0
        if pa.gpu:
            pa.gpu.push('au')
        call_data = []

        def cond(t, dt):
            call_data.append((t, dt))
            return False

        equations = [
            Group(
                equations=[
                    DumbEquation(dest='fluid', sources=['fluid'])
                ],
                condition=cond
            ),
            Group(
                equations=[
                    Group(
                        equations=[
                            DumbEquation(dest='fluid', sources=['fluid'])
                        ],
                        condition=cond
                    )
                ]
            ),
            Group(
                equations=[
                    DumbEquation(dest='fluid', sources=['fluid'])
                ],
            )
        ]
        a_eval = self._make_accel_eval(equations)

        # When
        a_eval.compute(0.0, 0.1)

        # Then
        if pa.gpu:
            pa.gpu.pull('au')

        expect = np.ones_like(pa.au)*7
        expect[0] = expect[-1] = 5
        expect[1] = expect[-2] = 6
        self.assertEqual(len(call_data), 2)
        self.assertEqual(call_data[0], (0.0, 0.1))
        self.assertEqual(call_data[1], (0.0, 0.1))
        self.assertListEqual(list(pa.au), list(expect))


class TestAccelerationEval1DGPUOctree(TestAccelerationEval1DGPU):
    def _get_nnps_cls(self):
        from pysph.base.gpu_nnps import OctreeGPUNNPS
        return OctreeGPUNNPS


class TestAccelerationEval1DGPUOctreeNonCached(
        TestAccelerationEval1DGPUOctree
):
    def setUp(self):
        self.old_flag = get_config().use_local_memory
        get_config().use_local_memory = True
        super(TestAccelerationEval1DGPUOctreeNonCached, self).setUp()

    def tearDown(self):
        super(TestAccelerationEval1DGPUOctreeNonCached, self).tearDown()
        get_config().use_local_memory = self.old_flag

    @pytest.mark.skip("Loop all not supported with non-cached NNPS")
    def test_should_support_loop_all_and_loop_on_gpu(self):
        pass

    @pytest.mark.skip("start/stop_idx not supported with non-cached NNPS")
    def test_should_honor_start_stop_idx_in_group_on_gpu(self):
        pass

    @pytest.mark.skip("start/stop_idx not supported with non-cached NNPS")
    def test_should_honor_start_stop_idx_as_str_in_group(self):
        pass


class TestAccelerationEval1DCUDA(TestAccelerationEval1DGPU):

    def _make_accel_eval(self, equations, cache_nnps=True):
        pytest.importorskip('pycuda')
        pytest.importorskip('pysph.base.gpu_nnps')
        GPUNNPS = self._get_nnps_cls()
        arrays = [self.pa]
        kernel = CubicSpline(dim=self.dim)
        a_eval = AccelerationEval(
            particle_arrays=arrays, equations=equations, kernel=kernel,
            backend='cuda'
        )
        comp = SPHCompiler(a_eval, integrator=None)
        comp.compile()
        self.sph_compiler = comp
        nnps = GPUNNPS(dim=kernel.dim, particles=arrays, cache=cache_nnps,
                       backend='cuda')
        nnps.update()
        a_eval.set_nnps(nnps)
        return a_eval

    def test_update_nnps_is_called_on_gpu(self):
        # Given
        equations = [
            Group(
                equations=[
                    SummationDensity(dest='fluid', sources=['fluid']),
                ],
                update_nnps=True
            ),
            Group(
                equations=[EqWithTime(dest='fluid', sources=['fluid'])]
            ),
        ]

        # When
        a_eval = self._make_accel_eval(equations)

        # Then
        h = a_eval.c_acceleration_eval.helper
        assert len(h.calls) == 5
        call = h.calls[0]
        assert call['type'] == 'kernel'
        assert call['loop'] is False

        call = h.calls[1]
        assert call['type'] == 'kernel'
        assert call['loop'] is True

        call = h.calls[2]
        assert call['type'] == 'method'
        assert call['method'] == 'update_nnps'

        call = h.calls[3]
        assert call['type'] == 'kernel'
        assert call['loop'] is False

        call = h.calls[4]
        assert call['type'] == 'kernel'
        assert call['loop'] is True


class TestAEvalMultipleDests(unittest.TestCase):
    def setUp(self):
        self.dim = 1
        n = 10
        dx = 1.0 / (n - 1)
        x = np.linspace(0, 1, n)
        m = np.ones_like(x)
        h = np.ones_like(x) * dx * 1.05
        self.pa1 = get_particle_array(name='f1', x=x, h=h, m=m)
        self.pa2 = get_particle_array(name='f2', x=x + dx/2, h=h, m=m)

    def _make_accel_eval(self, equations, cache_nnps=False):
        arrays = [self.pa1, self.pa2]
        kernel = CubicSpline(dim=self.dim)
        a_eval = AccelerationEval(
            particle_arrays=arrays, equations=equations, kernel=kernel
        )
        comp = SPHCompiler(a_eval, integrator=None)
        comp.compile()
        nnps = NNPS(dim=kernel.dim, particles=arrays, cache=cache_nnps)
        nnps.update()
        a_eval.set_nnps(nnps)
        return a_eval

    def test_update_nnps_should_only_be_called_once_per_group(self):
        # Given
        eqs = [
            SummationDensity(dest='f1', sources=['f1', 'f2']),
            SummationDensity(dest='f2', sources=['f1', 'f2']),
        ]
        equations = [Group(equations=eqs, update_nnps=True)]
        a_eval = self._make_accel_eval(equations)

        # This is quite ugly, it is not easy to mock out the set_nnps on an
        # acceleration eval as that requires a Cython object and not a mock
        # instance.
        profile_info = get_profile_info()
        n_up_dom = profile_info[0]['Integrator.update_domain']['calls']
        n_up = profile_info[0]['nnps.update']['calls']
        # When
        a_eval.compute(0.1, 0.1)

        # Then
        profile_info = get_profile_info()
        ncall = profile_info[0]['Integrator.update_domain']['calls']
        self.assertEqual(ncall, n_up_dom + 1)
        ncall = profile_info[0]['nnps.update']['calls']
        self.assertEqual(ncall, n_up + 1)


class TestAEvalMultipleDestsGPU(unittest.TestCase):
    def setUp(self):
        self.dim = 1
        n = 10
        dx = 1.0 / (n - 1)
        x = np.linspace(0, 1, n)
        m = np.ones_like(x)
        h = np.ones_like(x) * dx * 1.05
        self.pa1 = get_particle_array(name='f1', x=x, h=h, m=m)
        self.pa2 = get_particle_array(name='f2', x=x + dx/2, h=h, m=m)

    def _get_nnps_cls(self):
        from pysph.base.gpu_nnps import ZOrderGPUNNPS as GPUNNPS
        return GPUNNPS

    def _make_accel_eval(self, equations, cache_nnps=True):
        pytest.importorskip('pyopencl')
        pytest.importorskip('pysph.base.gpu_nnps')
        GPUNNPS = self._get_nnps_cls()
        arrays = [self.pa1, self.pa2]
        kernel = CubicSpline(dim=self.dim)
        a_eval = AccelerationEval(
            particle_arrays=arrays, equations=equations, kernel=kernel,
            backend='opencl'
        )
        comp = SPHCompiler(a_eval, integrator=None)
        comp.compile()
        self.sph_compiler = comp
        nnps = GPUNNPS(dim=kernel.dim, particles=arrays, cache=cache_nnps,
                       backend='opencl')
        nnps.update()
        a_eval.set_nnps(nnps)
        return a_eval

    def test_update_nnps_should_only_be_called_once_per_group(self):
        # Given
        eqs = [
            SummationDensity(dest='f1', sources=['f1', 'f2']),
            SummationDensity(dest='f2', sources=['f1', 'f2']),
        ]
        equations = [Group(equations=eqs, update_nnps=True)]

        # When
        a_eval = self._make_accel_eval(equations)

        # Then
        h = a_eval.c_acceleration_eval.helper
        update_nnps = [call for call in h.calls
                       if call['method'] == 'update_nnps']
        self.assertEqual(len(update_nnps), 1)



"""Test if dumping and loading files works in parallel correctly.
"""

import mpi4py.MPI as mpi
import numpy as np
from os.path import join
import shutil
from tempfile import mkdtemp

from pysph.base.particle_array import ParticleArray
from pysph.solver.utils import dump, load


def assert_lists_same(l1, l2):
    expect = list(sorted(l1))
    result = list(sorted(l2))
    assert expect == result, "Expected %s, got %s" % (expect, result)


def main():
    comm = mpi.COMM_WORLD
    rank = comm.Get_rank()
    size = comm.Get_size()

    root = mkdtemp()
    filename = join(root, 'test.npz')

    x = np.ones(5, dtype=float)*rank
    pa = ParticleArray(name='fluid', constants={'c1': 0.0, 'c2': [0.0, 0.0]},
                       x=x)

    try:
        dump(filename, [pa], {}, mpi_comm=comm)
        if rank == 0:
            data = load(filename)
            pa1 = data["arrays"]["fluid"]

            assert_lists_same(pa.properties.keys(), pa1.properties.keys())
            assert_lists_same(pa.constants.keys(), pa1.constants.keys())

            expect = np.ones(5*size)
            for i in range(size):
                expect[5*i:5*(i+1)] = i

            assert np.allclose(pa1.x, expect, atol=1e-14), \
                "Expected %s, got %s" % (expect, pa1.x)
    finally:
        shutil.rmtree(root)


if __name__ == '__main__':
    main()



# cython: language_level=3, embedsignature=True
# distutils: language=c++
from libcpp.vector cimport vector
from libcpp.map cimport map
from libcpp.pair cimport pair

from .nnps_base cimport *

cdef extern from 'math.h':
    int abs(int) nogil
    double ceil(double) nogil
    double floor(double) nogil
    double fabs(double) nogil
    double fmax(double, double) nogil
    double fmin(double, double) nogil

cdef extern from 'math.h':
    double log(double) nogil
    double log2(double) nogil

cdef extern from "z_order.h":
    ctypedef unsigned int uint32_t
    ctypedef unsigned long long uint64_t
    inline uint64_t get_key(uint64_t i, uint64_t j, uint64_t k) nogil

    cdef cppclass CompareSortWrapper:
        CompareSortWrapper() nogil except +
        CompareSortWrapper(uint32_t* current_pids, uint64_t* current_keys,
                int length) nogil except +
        inline void compare_sort() nogil

cdef class StratifiedSFCNNPS(NNPS):
    ############################################################################
    # Data Attributes
    ############################################################################
    cdef double radius_scale2
    cdef bint asymmetric

    cdef public int num_levels
    cdef int max_num_bits

    cdef uint64_t* max_keys
    cdef uint64_t current_max_key
    cdef uint64_t max_possible_key

    cdef double interval_size

    cdef uint32_t** pids
    cdef uint32_t* current_pids

    cdef uint64_t** keys
    cdef uint64_t* current_keys

    cdef int*** key_to_idx
    cdef int** current_key_to_idx

    cdef double** cell_sizes
    cdef double* current_cells

    cdef double*** hmax
    cdef double** current_hmax

    cdef int** num_cells
    cdef int* current_num_cells

    cdef vector[int]** nbr_boxes
    cdef vector[int]* current_nbr_boxes

    cdef int*** key_to_nbr_idx
    cdef int** current_key_to_nbr_idx

    cdef int*** key_to_nbr_length
    cdef int** current_key_to_nbr_length

    cdef int* total_mask_len
    cdef int current_mask_len

    cdef NNPSParticleArrayWrapper dst, src

    ##########################################################################
    # Member functions
    ##########################################################################
    cpdef np.ndarray get_keys(self, pa_index)

    cpdef set_context(self, int src_index, int dst_index)

    cdef inline int _get_H(self, double h_q, double h_j)

    cdef inline int get_idx(self, uint64_t key, uint64_t max_key, int* key_to_idx) nogil

    cdef void _fill_nbr_boxes(self)

    cdef int _neighbor_boxes_func(self, int i, int j, int k, int H,
            int* current_key_to_idx_level, uint64_t max_key,
            double current_cell_size, double* current_hmax_level, 
            vector[int]* nbr_boxes)

    cdef int _neighbor_boxes_asym(self, int i, int j, int k, int H,
            int* current_key_to_idx_level, uint64_t max_key,
            double current_cell_size, double* current_hmax_level,
            vector[int]* nbr_boxes) nogil

    cdef int _neighbor_boxes_sym(self, int i, int j, int k, int H,
            int* current_key_to_idx_level, uint64_t max_key,
            double current_cell_size, double* current_hmax_level,
            vector[int]* nbr_boxes) nogil

    cpdef double get_binning_size(self, int interval)

    cpdef int get_number_of_particles(self, int pa_index, int level)

    cpdef get_spatially_ordered_indices(self, int pa_index, LongArray indices)

    cdef void find_nearest_neighbors(self, size_t d_idx, UIntArray nbrs) nogil

    cdef void fill_array(self, NNPSParticleArrayWrapper pa_wrapper,
            int pa_index, uint32_t* current_pids,
            uint64_t* current_keys, double* current_cells, int* current_num_cells)

    cdef inline int _get_level(self, double h) nogil

    cpdef _refresh(self)

    cpdef _bin(self, int pa_index, UIntArray indices)



from __future__ import print_function
import logging
import numpy as np
from pytools import memoize, memoize_method
import mako.template as mkt

from compyle.config import get_config
from compyle.array import get_backend, wrap_array, Array
from compyle.parallel import Elementwise, Scan
from compyle.api import declare, annotate
from compyle.types import dtype_to_ctype
from compyle.template import Template
import compyle.array as array
import pysph.base.particle_array


logger = logging.getLogger()


class ExtractParticles(Template):
    def __init__(self, name, prop_names):
        super(ExtractParticles, self).__init__(name=name)
        self.prop_names = prop_names

    def extra_args(self):
        args = []
        for prop in self.prop_names:
            args.append('stride_%s' % prop)
            args.append('dst_%s' % prop)
            args.append('src_%s' % prop)
        return args, {}

    def template(self, i, indices, start_idx):
        '''
        idx, s_idx, s_i, j, start = declare('int', 5)
        idx = indices[i]
        % for prop in obj.prop_names:
        s_idx = stride_${prop} * idx
        s_i = stride_${prop} * i
        start = stride_${prop} * start_idx
        for j in range(stride_${prop}):
            dst_${prop}[start + s_i + j] = src_${prop}[s_idx + j]
        % endfor
        '''


class DeviceHelper(object):
    """Manages the arrays contained in a particle array on the device.

    Note that it converts the data to a suitable type depending on the value of
    get_config().use_double. Further, note that it assumes that the names of
    constants and properties do not clash.

    """

    def __init__(self, particle_array, backend=None):
        self.backend = get_backend(backend)
        self._particle_array = pa = particle_array
        self.use_double = get_config().use_double
        self._dtype = np.float64 if self.use_double else np.float32
        self.num_real_particles = pa.num_real_particles
        self._data = {}
        self.properties = []
        self.constants = []
        self._strided_indices_knl = None

        for prop, ary in pa.properties.items():
            self.add_prop(prop, ary)
        for prop, ary in pa.constants.items():
            self.add_const(prop, ary)

    def _get_array(self, ary):
        ctype = ary.get_c_type()
        if ctype in ['float', 'double']:
            return ary.get_npy_array().astype(self._dtype)
        else:
            return ary.get_npy_array()

    def _get_prop_or_const(self, prop):
        pa = self._particle_array
        return pa.properties.get(prop, pa.constants.get(prop))

    def _add_prop_or_const(self, name, carray):
        """Add a new property or constant given the name and carray, note
        that this assumes that this property is already added to the
        particle array.
        """
        np_array = self._get_array(carray)
        g_ary = Array(np_array.dtype, n=carray.length,
                      backend=self.backend)
        g_ary.set(np_array)
        self._data[name] = g_ary
        setattr(self, name, g_ary)

    def get_number_of_particles(self, real=False):
        if real:
            return self.num_real_particles
        else:
            if len(self.properties) > 0:
                pname = self.properties[0]
                stride = self._particle_array.stride.get(pname, 1)
                prop0 = self._data[pname]
                return len(prop0.dev) // stride
            else:
                return 0

    def align(self, indices):
        '''Note that the indices passed here is a dictionary
        keyed on the stride.
        '''
        if not isinstance(indices, dict):
            indices = {1: indices}
            self._make_strided_indices(indices)

        for prop in self.properties:
            stride = self._particle_array.stride.get(prop, 1)
            self._data[prop] = self._data[prop].align(indices.get(stride))
            setattr(self, prop, self._data[prop])

    def _make_strided_indices(self, indices_dict):
        '''Takes the indices in a dict assuming that the indices are for a
        stride of 1 and makes suitable indices for other possible stride
        values.
        '''
        indices = indices_dict[1]
        n = indices.length
        if not self._strided_indices_knl:
            self._setup_strided_indices_kernel()
        for stride in set(self._particle_array.stride.values()):
            dest = array.empty(n*stride, dtype=np.int32, backend=self.backend)
            self._strided_indices_knl(indices, dest, stride)
            indices_dict[stride] = dest

    def _setup_strided_indices_kernel(self):
        @annotate(int='i, stride', gintp='indices, dest')
        def set_indices(i, indices, dest, stride):
            j = declare('int')
            for j in range(stride):
                dest[i*stride + j] = indices[i]*stride + j

        knl = Elementwise(set_indices, backend=self.backend)
        self._strided_indices_knl = knl

    def add_prop(self, name, carray):
        """Add a new property given the name and carray, note
        that this assumes that this property is already added to the
        particle array.
        """
        if name in self._particle_array.properties and \
                name not in self.properties:
            self._add_prop_or_const(name, carray)
            self.properties.append(name)

    def add_const(self, name, carray):
        """Add a new constant given the name and carray, note
        that this assumes that this property is already added to the
        particle array.
        """
        if name in self._particle_array.constants and \
                name not in self.constants:
            self._add_prop_or_const(name, carray)
            self.constants.append(name)

    def update_prop(self, name, dev_array):
        """Add a new property to DeviceHelper. Note that this property
        is not added to the particle array itself"""
        self._data[name] = dev_array
        setattr(self, name, dev_array)
        if name not in self.properties:
            self.properties.append(name)

    def update_const(self, name, dev_array):
        """Add a new constant to DeviceHelper. Note that this property
        is not added to the particle array itself"""
        self._data[name] = dev_array
        setattr(self, name, dev_array)
        if name not in self.constants:
            self.constants.append(name)

    def get_device_array(self, prop):
        if prop in self.properties or prop in self.constants:
            return self._data[prop]

    def max(self, arg):
        return float(array.maximum(getattr(self, arg),
                                   backend=self.backend))

    def update_minmax_cl(self, props, only_min=False, only_max=False):
        ary_list = [getattr(self, prop) for prop in props]
        array.update_minmax_gpu(ary_list, only_min=only_min,
                                only_max=only_max, backend=self.backend)

    def update_min_max(self, props=None):
        """Update the min,max values of all properties """
        props = props if props else self.properties
        for prop in props:
            array = self._data[prop]
            array.update_min_max()

    def pull(self, *args):
        p = self._particle_array
        if len(args) == 0:
            args = self._data.keys()
        for arg in args:
            arg_cpu = getattr(self, arg).get()
            if arg in p.properties or arg in p.constants:
                pa_arr = self._get_prop_or_const(arg)
            else:
                if arg in self.properties:
                    p.add_property(arg)
                if arg in self.constants:
                    p.add_constant(arg, arg_cpu)
                pa_arr = self._get_prop_or_const(arg)
            if arg_cpu.size != pa_arr.length:
                pa_arr.resize(arg_cpu.size)
            pa_arr.set_data(arg_cpu)
        p.set_num_real_particles(self.num_real_particles)

    def push(self, *args):
        if len(args) == 0:
            args = self._data.keys()
        for arg in args:
            dev_arr = array.to_device(
                self._get_array(self._get_prop_or_const(arg)),
                backend=self.backend)
            self._data[arg].set_data(dev_arr)
            setattr(self, arg, self._data[arg])

    def _check_property(self, prop):
        """Check if a property is present or not """
        if prop in self.properties or prop in self.constants:
            return
        else:
            raise AttributeError('property %s not present' % (prop))

    def remove_prop(self, name):
        if name in self.properties:
            self.properties.remove(name)
        if name in self._data:
            del self._data[name]
            delattr(self, name)

    def resize(self, new_size):
        for prop in self.properties:
            stride = self._particle_array.stride.get(prop, 1)
            self._data[prop].resize(new_size * stride)
            setattr(self, prop, self._data[prop])

    @memoize_method
    def _get_align_kernel_without_strides(self):
        @annotate(i='int', tag_arr='gintp', return_='int')
        def align_input_expr(i, tag_arr):
            return tag_arr[i] == 0

        @annotate(int='i, item, prev_item, last_item, num_particles',
                  gintp='tag_arr, new_indices, num_real_particles')
        def align_output_expr(i, item, prev_item, last_item, tag_arr,
                              new_indices, num_particles, num_real_particles):
            t, idx = declare('int', 2)
            t = last_item + i - prev_item
            idx = t if tag_arr[i] else prev_item
            new_indices[idx] = i
            if i == num_particles - 1:
                num_real_particles[0] = last_item

        align_particles_knl = Scan(align_input_expr, align_output_expr,
                                   'a+b', dtype=np.int32, backend=self.backend)

        return align_particles_knl

    @memoize_method
    def _get_align_kernel_with_strides(self):
        @annotate(i='int', tag_arr='gintp', return_='int')
        def align_input_expr(i, tag_arr):
            return tag_arr[i] == 0

        @annotate(int='i, item, prev_item, last_item, stride, num_particles',
                  gintp='tag_arr, new_indices',
                  return_='int')
        def align_output_expr(i, item, prev_item, last_item, tag_arr,
                              new_indices, num_particles, stride):
            t, idx, j_s = declare('int', 3)
            t = last_item + i - prev_item
            idx = t if tag_arr[i] else prev_item
            for j_s in range(stride):
                new_indices[stride * idx + j_s] = stride * i + j_s

        align_particles_knl = Scan(align_input_expr, align_output_expr,
                                   'a+b', dtype=np.int32, backend=self.backend)

        return align_particles_knl

    def align_particles(self):
        tag_arr = self._data['tag']

        if len(tag_arr) == 0:
            self.num_real_particles = 0
            return

        num_particles = len(tag_arr)

        new_indices = array.empty(num_particles, dtype=np.int32,
                                  backend=self.backend)
        num_real_particles = array.empty(1, dtype=np.int32,
                                         backend=self.backend)

        align_particles_knl = self._get_align_kernel_without_strides()

        align_particles_knl(tag_arr=tag_arr, new_indices=new_indices,
                            num_particles=num_particles,
                            num_real_particles=num_real_particles)

        indices = {1: new_indices}
        for stride in set(self._particle_array.stride.values()):
            if stride > 1:
                indices[stride] = self._build_indices_with_strides(
                    tag_arr, stride
                )

        self.align(indices)

        self.num_real_particles = int(num_real_particles.get())

    def _build_indices_with_strides(self, tag_arr, stride):
        num_particles = len(tag_arr.dev)

        new_indices = array.empty(num_particles * stride,
                                  dtype=np.int32,
                                  backend=self.backend)

        align_particles_knl = self._get_align_kernel_with_strides()

        align_particles_knl(tag_arr=tag_arr, new_indices=new_indices,
                            num_particles=num_particles,
                            stride=stride)

        return new_indices

    @memoize_method
    def _get_remove_particles_bool_kernels(self):
        @annotate(i='int', if_remove='gintp', return_='int')
        def remove_input_expr(i, if_remove):
            return if_remove[i]

        @annotate(int='i, item, last_item, num_particles',
                  gintp='if_remove, num_removed_particles',
                  new_indices='guintp')
        def remove_output_expr(i, item, last_item, if_remove, new_indices,
                               num_removed_particles, num_particles):
            if not if_remove[i]:
                new_indices[i - item] = i
            if i == num_particles - 1:
                num_removed_particles[0] = last_item

        remove_knl = Scan(remove_input_expr, remove_output_expr,
                          'a+b', dtype=np.int32, backend=self.backend)

        @annotate(int='i, size, stride', guintp='indices, new_indices')
        def stride_knl_elwise(i, indices, new_indices, size, stride):
            tmp_idx, j_s = declare('unsigned int', 2)
            for j_s in range(stride):
                tmp_idx = i * stride + j_s
                if tmp_idx < size:
                    new_indices[tmp_idx] = indices[i] * stride + j_s

        stride_knl = Elementwise(stride_knl_elwise, backend=self.backend)

        return remove_knl, stride_knl

    def _remove_particles_bool(self, if_remove, align=True):
        """ Remove particle i if if_remove[i] is True
        """
        num_indices = int(array.sum(if_remove, backend=self.backend))

        if num_indices == 0:
            return

        num_particles = self.get_number_of_particles()
        new_indices = Array(np.uint32, n=(num_particles - num_indices),
                            backend=self.backend)
        num_removed_particles = array.empty(1, dtype=np.int32,
                                            backend=self.backend)

        remove_knl, stride_knl = self._get_remove_particles_bool_kernels()

        remove_knl(if_remove=if_remove, new_indices=new_indices,
                   num_removed_particles=num_removed_particles,
                   num_particles=num_particles)

        new_num_particles = num_particles - int(num_removed_particles.get())

        strides = set(self._particle_array.stride.values())
        s_indices = {1: new_indices}
        for stride in strides:
            if stride == 1:
                continue
            size = new_num_particles * stride
            s_index = Array(np.uint32, n=size, backend=self.backend)
            stride_knl(new_indices, s_index, size, stride)
            s_indices[stride] = s_index

        for prop in self.properties:
            stride = self._particle_array.stride.get(prop, 1)
            s_index = s_indices[stride]
            self._data[prop] = self._data[prop].align(s_index)
            setattr(self, prop, self._data[prop])

        if align:
            self.align_particles()

    @memoize_method
    def _get_remove_particles_kernel(self):
        @annotate(int='i, size', indices='guintp', if_remove='gintp')
        def fill_if_remove_elwise(i, indices, if_remove, size):
            if indices[i] < size:
                if_remove[indices[i]] = 1

        fill_if_remove_knl = Elementwise(fill_if_remove_elwise,
                                         backend=self.backend)

        return fill_if_remove_knl

    def remove_particles(self, indices, align=True):
        """ Remove particles whose indices are given in index_list.

        Parameters
        ----------

        indices : array
            an array of indices, this array can be a list, numpy array
            or a LongArray.

        """
        if len(indices) > self.get_number_of_particles():
            msg = 'Number of particles to be removed is greater than'
            msg += 'number of particles in array'
            raise ValueError(msg)

        num_particles = self.get_number_of_particles()
        if_remove = Array(np.int32, n=num_particles, backend=self.backend)
        if_remove.fill(0)

        fill_if_remove_knl = self._get_remove_particles_kernel()
        fill_if_remove_knl(indices, if_remove, num_particles)

        self._remove_particles_bool(if_remove, align=align)

    def remove_tagged_particles(self, tag, align=True):
        """ Remove particles that have the given tag.

        Parameters
        ----------

        tag : int
            the type of particles that need to be removed.

        """
        tag_array = self.tag
        if_remove = wrap_array((tag_array.dev == tag).astype(np.int32),
                               self.backend)
        self._remove_particles_bool(if_remove, align=align)

    def add_particles(self, align=True, **particle_props):
        """
        Add particles in particle_array to self.

        Parameters
        ----------

        particle_props : dict
            a dictionary containing cl arrays for various particle
            properties.

        Notes
        -----

         - all properties should have same length arrays.
         - all properties should already be present in this particles array.
           if new properties are seen, an exception will be raised.
           properties.

        """
        pa = self._particle_array

        if len(particle_props) == 0:
            return

        # check if the input properties are valid.
        for prop in particle_props:
            self._check_property(prop)

        num_extra_particles = len(list(particle_props.values())[0])
        old_num_particles = self.get_number_of_particles()
        new_num_particles = num_extra_particles + old_num_particles

        for prop in self.properties:
            arr = self._data[prop]
            stride = self._particle_array.stride.get(prop, 1)
            if prop in particle_props.keys():
                s_arr = particle_props[prop]
                arr.extend(s_arr)
            else:
                arr.resize(new_num_particles * stride)
                # set the properties of the new particles to the default ones.
                arr[old_num_particles * stride:] = pa.default_values[prop]

            self.update_prop(prop, arr)

        if num_extra_particles > 0 and align:
            # make sure particles are aligned properly.
            self.align_particles()

        return 0

    def extend(self, num_particles):
        """ Increase the total number of particles by the requested amount

        New particles are added at the end of the list, you may
        have to manually call align_particles later.
        """
        if num_particles <= 0:
            return

        old_size = self.get_number_of_particles()
        new_size = old_size + num_particles

        for prop in self.properties:
            arr = self._data[prop]
            stride = self._particle_array.stride.get(prop, 1)
            arr.resize(new_size * stride)
            arr[old_size * stride:] = \
                self._particle_array.default_values[prop]
            self.update_prop(prop, arr)

    def append_parray(self, parray, align=True, update_constants=False):
        """ Add particles from a particle array

        properties that are not there in self will be added
        """
        if parray.gpu is None:
            parray.set_device_helper(DeviceHelper(parray))

        if parray.gpu.get_number_of_particles() == 0:
            return

        num_extra_particles = parray.gpu.get_number_of_particles()
        old_num_particles = self.get_number_of_particles()
        new_num_particles = num_extra_particles + old_num_particles

        # extend current arrays by the required number of particles
        self.extend(num_extra_particles)

        my_stride = self._particle_array.stride
        for prop_name in parray.gpu.properties:
            stride = parray.stride.get(prop_name, 1)
            if stride > 1 and prop_name not in my_stride:
                my_stride[prop_name] = stride
            if prop_name in self.properties:
                arr = self._data[prop_name]
                source = parray.gpu.get_device_array(prop_name)
                arr.dev[old_num_particles * stride:] = source.dev
            else:
                # meaning this property is not there in self.
                dtype = parray.gpu.get_device_array(prop_name).dtype
                arr = Array(dtype, n=new_num_particles * stride,
                            backend=self.backend)
                arr.fill(parray.default_values[prop_name])
                self.update_prop(prop_name, arr)

                # now add the values to the end of the created array
                dest = self._data[prop_name]
                source = parray.gpu.get_device_array(prop_name)
                dest.dev[old_num_particles * stride:] = source.dev

        if update_constants:
            for const in parray.gpu.constants:
                if const not in self.constants:
                    arr = parray.gpu.get_device_array(const)
                    self.update_const(const, arr.copy())

        if num_extra_particles > 0 and align:
            self.align_particles()

    def empty_clone(self, props=None):
        if props is None:
            prop_names = self.properties
        else:
            prop_names = props

        result_array = pysph.base.particle_array.ParticleArray(
            backend=self._particle_array.backend
        )
        result_array.set_name(self._particle_array.name)
        result_array.set_device_helper(DeviceHelper(result_array,
                                                    backend=self.backend))

        for prop_name in prop_names:
            src_arr = self._data[prop_name]
            stride = self._particle_array.stride.get(prop_name, 1)
            prop_type = dtype_to_ctype(src_arr.dtype)
            prop_default = self._particle_array.default_values[prop_name]
            result_array.add_property(name=prop_name,
                                      type='double' if prop_type == 'float' else prop_type,
                                      default=prop_default,
                                      stride=stride)

        for const in self.constants:
            result_array.gpu.update_const(const, self._data[const].copy())

        if props is None:
            output_arrays = list(self._particle_array.output_property_arrays)
        else:
            output_arrays = list(
                set(props).intersection(
                    self._particle_array.output_property_arrays
                )
            )

        result_array.set_output_arrays(output_arrays)
        return result_array

    def extract_particles(self, indices, dest_array=None,
                          align=True, props=None):
        """Create new particle array for particles with given indices

        Parameters
        ----------

        indices : Array
            indices of particles to be extracted.

        props : list
            the list of properties to extract, if None all properties
            are extracted.

        """
        if not dest_array:
            dest_array = self.empty_clone(props=props)

        if props is None:
            prop_names = self.properties
        else:
            prop_names = props

        if len(indices) == 0:
            return dest_array

        start_idx = dest_array.get_number_of_particles()

        dest_array.gpu.extend(len(indices))

        args_list = [indices, start_idx]

        for prop in prop_names:
            stride = self._particle_array.stride.get(prop, 1)
            src_arr = self._data[prop]
            dst_arr = dest_array.gpu.get_device_array(prop)

            args_list.append(stride)
            args_list.append(dst_arr)
            args_list.append(src_arr)

        extract_particles_knl = ExtractParticles('extract_particles_knl',
                                                 prop_names).function
        extract_particles_elwise = Elementwise(extract_particles_knl,
                                               backend=self.backend)
        extract_particles_elwise(*args_list)

        if align:
            dest_array.gpu.align_particles()

        return dest_array



# cython: language_level=3, embedsignature=True
# distutils: language=c++

# malloc and friends
from libc.stdlib cimport malloc, free
from libcpp.vector cimport vector
from libcpp.map cimport map

from cython.operator cimport dereference as deref, preincrement as inc

# Cython for compiler directives
cimport cython

cdef extern from "<algorithm>" namespace "std" nogil:
    void sort[Iter, Compare](Iter first, Iter last, Compare comp)
    void sort[Iter](Iter first, Iter last)

IF UNAME_SYSNAME == "Windows":
    @cython.cdivision(True)
    cdef inline double log2(double n) nogil:
        return log(n)/log(2)

#############################################################################

cdef class CellIndexingNNPS(NNPS):

    """Find nearest neighbors using cell indexing.

    Uses a sorted array to access particles in a cell. Gives better cache performance
    than spatial hash.

    Ref: J. Onderik, R. Durikovic, Efficient Neighbor Search for Particle-based Fluids,
    Journal of the Applied Mathematics, Statistics and Informatics 4 (1) (2008) 29–43.
    http://citeseerx.ist.psu.edu/viewdoc/download?doi=10.1.1.105.6732&rep=rep1&type=pdf
    """

    def __init__(self, int dim, list particles, double radius_scale = 2.0,
            int ghost_layers = 1, domain=None, bint fixed_h = False,
            bint cache = False, bint sort_gids = False):
        NNPS.__init__(
            self, dim, particles, radius_scale, ghost_layers, domain,
            cache, sort_gids
        )

        self.radius_scale2 = radius_scale*radius_scale
        cdef NNPSParticleArrayWrapper pa_wrapper
        cdef int i, num_particles

        for i from 0<=i<self.narrays:
            pa_wrapper = <NNPSParticleArrayWrapper> self.pa_wrappers[i]
            num_particles = pa_wrapper.get_number_of_particles()

            self.keys[i] = <u_int*> malloc(num_particles*sizeof(u_int))
            self.key_indices[i] = new key_to_idx_t()

        self.src_index = 0
        self.dst_index = 0
        self.sort_gids = sort_gids
        self.domain.update()
        self.update()

    def __cinit__(self, int dim, list particles, double radius_scale = 2.0,
            int ghost_layers = 1, domain=None, bint fixed_h = False,
            bint cache = False, bint sort_gids = False):
        cdef int narrays = len(particles)

        self.keys = <u_int**> malloc(narrays*sizeof(u_int*))
        self.key_indices = <key_to_idx_t**> malloc(narrays*sizeof(key_to_idx_t*))

        self.I = <u_int*> malloc(narrays*sizeof(u_int))

        self.current_keys = NULL
        self.current_indices = NULL

    def __dealloc__(self):
        cdef int i
        for i from 0<=i<self.narrays:
            free(self.keys[i])
            del self.key_indices[i]
        free(self.keys)
        free(self.key_indices)
        free(self.I)


    #### Public protocol ################################################

    cpdef set_context(self, int src_index, int dst_index):
        """Set context for nearest neighbor searches.

        Parameters
        ----------
        src_index: int
            Index in the list of particle arrays to which the neighbors belong

        dst_index: int
            Index in the list of particle arrays to which the query point belongs

        """
        NNPS.set_context(self, src_index, dst_index)
        self.current_keys = self.keys[src_index]
        self.current_indices = self.key_indices[src_index]

        self.dst = <NNPSParticleArrayWrapper> self.pa_wrappers[dst_index]
        self.src = <NNPSParticleArrayWrapper> self.pa_wrappers[src_index]

    cdef void find_nearest_neighbors(self, size_t d_idx, UIntArray nbrs) nogil:
        """Low level, high-performance non-gil method to find neighbors.
        This requires that `set_context()` be called beforehand.  This method
        does not reset the neighbors array before it appends the
        neighbors to it.

        """
        cdef double* dst_x_ptr = self.dst.x.data
        cdef double* dst_y_ptr = self.dst.y.data
        cdef double* dst_z_ptr = self.dst.z.data
        cdef double* dst_h_ptr = self.dst.h.data

        cdef double* src_x_ptr = self.src.x.data
        cdef double* src_y_ptr = self.src.y.data
        cdef double* src_z_ptr = self.src.z.data
        cdef double* src_h_ptr = self.src.h.data

        cdef double x = dst_x_ptr[d_idx]
        cdef double y = dst_y_ptr[d_idx]
        cdef double z = dst_z_ptr[d_idx]
        cdef double h = dst_h_ptr[d_idx]

        cdef unsigned int* s_gid = self.src.gid.data
        cdef int orig_length = nbrs.length

        cdef int c_x, c_y, c_z
        cdef double* xmin = self.xmin.data
        cdef int i, j

        find_cell_id_raw(
                x - xmin[0],
                y - xmin[1],
                z - xmin[2],
                self.cell_size,
                &c_x, &c_y, &c_z
                )

        cdef double xij2 = 0
        cdef double hi2 = self.radius_scale2*h*h
        cdef double hj2 = 0

        cdef map[u_int, pair[u_int, u_int]].iterator it

        cdef int x_boxes[27]
        cdef int y_boxes[27]
        cdef int z_boxes[27]
        cdef int num_boxes = self._neighbor_boxes(c_x, c_y, c_z,
                x_boxes, y_boxes, z_boxes)

        cdef pair[u_int, u_int] candidate

        cdef u_int n, idx
        for i from 0<=i<num_boxes:
            it = self.current_indices.find(self._get_key(0, x_boxes[i], y_boxes[i],
                z_boxes[i], self.src_index))
            if it == self.current_indices.end():
                continue
            candidate = deref(it).second
            n = candidate.first
            candidate_length = candidate.second

            for j from 0<=j<candidate_length:
                idx = self._get_id(self.current_keys[n+j], self.src_index)

                hj2 = self.radius_scale2*src_h_ptr[idx]*src_h_ptr[idx]

                xij2 = norm2(
                    src_x_ptr[idx] - x,
                    src_y_ptr[idx] - y,
                    src_z_ptr[idx] - z
                    )

                if (xij2 < hi2) or (xij2 < hj2):
                    nbrs.c_append(idx)

        if self.sort_gids:
            self._sort_neighbors(
                &nbrs.data[orig_length], nbrs.length - orig_length, s_gid
            )


    cpdef get_nearest_particles_no_cache(self, int src_index, int dst_index,
            size_t d_idx, UIntArray nbrs, bint prealloc):
        """Find nearest neighbors for particle id 'd_idx' without cache

        Parameters
        ----------
        src_index: int
            Index in the list of particle arrays to which the neighbors belong

        dst_index: int
            Index in the list of particle arrays to which the query point belongs

        d_idx: size_t
            Index of the query point in the destination particle array

        nbrs: UIntArray
            Array to be populated by nearest neighbors of 'd_idx'

        """
        self.set_context(src_index, dst_index)

        if prealloc:
            nbrs.length = 0
        else:
            nbrs.c_reset()

        self.find_nearest_neighbors(d_idx, nbrs)

    cpdef get_spatially_ordered_indices(self, int pa_index, LongArray indices):
        indices.reset()
        cdef NNPSParticleArrayWrapper pa_wrapper = self.pa_wrappers[pa_index]
        cdef int num_particles = pa_wrapper.get_number_of_particles()

        cdef u_int* current_keys = self.keys[pa_index]

        cdef int j
        for j from 0<=j<num_particles:
            indices.c_append(<long>self._get_id(current_keys[j], pa_index))


    cdef void fill_array(self, NNPSParticleArrayWrapper pa_wrapper, int pa_index,
            UIntArray indices, u_int* current_keys, key_to_idx_t* current_indices) nogil:
        cdef double* x_ptr = pa_wrapper.x.data
        cdef double* y_ptr = pa_wrapper.y.data
        cdef double* z_ptr = pa_wrapper.z.data

        cdef double* xmin = self.xmin.data

        cdef int i, n
        cdef int c_x, c_y, c_z
        for i from 0<=i<indices.length:
            n = indices.data[i]
            find_cell_id_raw(
                    x_ptr[i] - xmin[0],
                    y_ptr[i] - xmin[1],
                    z_ptr[i] - xmin[2],
                    self.cell_size,
                    &c_x, &c_y, &c_z
                    )
            current_keys[i] = self._get_key(n, c_x, c_y, c_z, pa_index)

        sort(current_keys, current_keys + indices.length)

        cdef int id_x, id_y, id_z

        cdef pair[u_int, pair[u_int, u_int]] temp
        cdef pair[u_int, u_int] cell

        c_x = self._get_x(current_keys[0], pa_index)
        c_y = self._get_y(current_keys[0], pa_index)
        c_z = self._get_z(current_keys[0], pa_index)

        temp.first = self._get_key(0, c_x, c_y, c_z, pa_index)
        cell.first = 0

        cdef u_int length = 0

        for i from 0<i<indices.length:
            id_x = self._get_x(current_keys[i], pa_index)
            id_y = self._get_y(current_keys[i], pa_index)
            id_z = self._get_z(current_keys[i], pa_index)

            length += 1

            if(id_x != c_x or id_y != c_y or id_z != c_z):
                cell.second = length
                temp.second = cell
                current_indices.insert(temp)

                temp.first = self._get_key(0, id_x, id_y, id_z, pa_index)
                cell.first = i

                length = 0

                c_x = id_x
                c_y = id_y
                c_z = id_z

        cell.second = length + 1
        temp.second = cell
        current_indices.insert(temp)

    #### Private protocol ################################################

    cdef inline u_int _get_key(self, u_int n, u_int i, u_int j,
            u_int k, int pa_index) nogil:
        return  n + \
                (1 << self.I[pa_index])*i + \
                (1 << (self.I[pa_index] + self.J))*j + \
                (1 << (self.I[pa_index] + self.J + self.K))*k

    @cython.cdivision(True)
    cdef inline int _get_id(self, u_int key, int pa_index) nogil:
        return key % (1 << self.I[pa_index])

    @cython.cdivision(True)
    cdef inline int _get_x(self, u_int key, int pa_index) nogil:
        return (key >> self.I[pa_index]) % (1 << self.J)

    @cython.cdivision(True)
    cdef inline int _get_y(self, u_int key, int pa_index) nogil:
        return (key >> (self.I[pa_index] + self.J)) % (1 << self.K)

    @cython.cdivision(True)
    cdef inline int _get_z(self, u_int key, int pa_index) nogil:
        return key >> (self.I[pa_index] + self.J + self.K)

    cdef inline int _neighbor_boxes(self, int i, int j, int k,
            int* x, int* y, int* z) nogil:
        cdef int length = 0
        cdef int p, q, r
        for p from -1<=p<2:
            for q from -1<=q<2:
                for r from -1<=r<2:
                    if i+r>=0 and j+q>=0 and k+p>=0:
                        x[length] = i+r
                        y[length] = j+q
                        z[length] = k+p
                        length += 1
        return length

    cpdef _refresh(self):
        cdef NNPSParticleArrayWrapper pa_wrapper

        # Only necessary if number of particles in a ParticleArray changes
        cdef int i, num_particles

        cdef double* xmax = self.xmax.data
        cdef double* xmin = self.xmin.data

        self.J = <u_int> (1 + log2(ceil((xmax[0] - xmin[0])/self.cell_size)))
        self.K = <u_int> (1 + log2(ceil((xmax[1] - xmin[1])/self.cell_size)))

        for i from 0<=i<self.narrays:
            free(self.keys[i])
            del self.key_indices[i]

            pa_wrapper = <NNPSParticleArrayWrapper> self.pa_wrappers[i]
            num_particles = pa_wrapper.get_number_of_particles()

            self.keys[i] = <u_int*> malloc(num_particles*sizeof(u_int))
            self.key_indices[i] = new key_to_idx_t()

        self.current_keys = self.keys[self.src_index]
        self.current_indices = self.key_indices[self.src_index]

    @cython.cdivision(True)
    cpdef _bin(self, int pa_index, UIntArray indices):
        cdef NNPSParticleArrayWrapper pa_wrapper = self.pa_wrappers[pa_index]
        cdef int num_particles = pa_wrapper.get_number_of_particles()

        self.I[pa_index] = <u_int> (1 + log2(pa_wrapper.get_number_of_particles()))

        cdef u_int* current_keys = self.keys[pa_index]
        cdef key_to_idx_t* current_indices = self.key_indices[pa_index]

        self.fill_array(pa_wrapper, pa_index, indices, current_keys, current_indices)



""" This is a setup example that will be used other gas_dynamics problem
    like SodShockTube, BlastWave
"""

import os
import numpy
from math import sqrt

from pysph.base.utils import get_particle_array as gpa
from pysph.solver.application import Application
from pysph.examples.gas_dynamics import riemann_solver


class ShockTubeSetup(Application):

    def generate_particles(self, xmin, xmax, dxl, dxr, m, pl, pr, h0,  bx,
                           gamma1, ul=0, ur=0, constants={}):
        xt1 = numpy.arange(xmin - bx + 0.5 * dxl, 0, dxl)
        xt2 = numpy.arange(0.5 * dxr, xmax + bx, dxr)
        xt = numpy.concatenate([xt1, xt2])
        leftb_indices = numpy.where(xt <= xmin)[0]
        left_indices = numpy.where((xt > xmin) & (xt < 0))[0]
        right_indices = numpy.where((xt >= 0) & (xt < xmax))[0]
        rightb_indices = numpy.where(xt >= xmax)[0]
        x1 = xt[left_indices]
        x2 = xt[right_indices]
        b1 = xt[leftb_indices]
        b2 = xt[rightb_indices]

        x = numpy.concatenate([x1, x2])
        b = numpy.concatenate([b1, b2])
        right_indices = numpy.where(x > 0.0)[0]

        rho = numpy.ones_like(x) * m / dxl
        rho[right_indices] = m / dxr

        p = numpy.ones_like(x) * pl
        p[right_indices] = pr

        u = numpy.ones_like(x) * ul
        u[right_indices] = ur

        h = numpy.ones_like(x) * h0
        m = numpy.ones_like(x) * m
        e = p / (gamma1 * rho)
        wij = numpy.ones_like(x)

        bwij = numpy.ones_like(b)
        brho = numpy.ones_like(b)
        bp = numpy.ones_like(b)
        be = bp / (gamma1 * brho)
        bm = numpy.ones_like(b) * dxl
        bh = numpy.ones_like(b) * 4 * h0
        bhtmp = numpy.ones_like(b)
        fluid = gpa(
            constants=constants, name='fluid', x=x, rho=rho, p=p,
            e=e, h=h, m=m, u=u, wij=wij, h0=h.copy()
        )

        boundary = gpa(
            constants=constants, name='boundary', x=b, rho=brho, p=bp,
            e=be, h=bh, m=bm, wij=bwij, h0=bh.copy(), htmp=bhtmp
        )

        self.scheme.setup_properties([fluid, boundary])
        print("1D Shocktube with %d particles" %
              (fluid.get_number_of_particles()))
        return [fluid, boundary]

    def post_process(self):
        try:
            import matplotlib
            matplotlib.use('Agg')
            from matplotlib import pyplot as plt
        except ImportError:
            print("Post processing requires matplotlib.")
            return
        if self.rank > 0 or len(self.output_files) == 0:
            return

        last_output = self.output_files[-1]
        from pysph.solver.utils import load
        data = load(last_output)
        pa = data['arrays']['fluid']
        gamma = self.options.gamma if self.options.gamma else 1.4
        riemann_solver.set_gamma(gamma)

        rho_e, u_e, p_e, e_e, x_e = riemann_solver.solve(
            x_min=self.xmin, x_max=self.xmax, x_0=self.x0,
            t=self.tf, p_l=self.pl, p_r=self.pr, rho_l=self.rhol,
            rho_r=self.rhor, u_l=self.ul, u_r=self.ur, N=101
        )
        x = pa.x
        rho = pa.rho
        e = pa.e
        cs = pa.cs
        u = pa.u
        p = pa.p
        h = pa.h

        plt.plot(x, rho, label='pysph (' + str(self.options.scheme) + ')')
        plt.plot(x_e, rho_e, label='exact')
        plt.xlabel('x')
        plt.ylabel('rho')
        plt.legend()
        fig = os.path.join(self.output_dir, "density.png")
        plt.savefig(fig, dpi=300)
        plt.clf()
        plt.plot(x, e, label='pysph (' + str(self.options.scheme) + ')')
        plt.plot(x_e, e_e, label='exact')
        plt.xlabel('x')
        plt.ylabel('e')
        plt.legend()
        fig = os.path.join(self.output_dir, "energy.png")
        plt.savefig(fig, dpi=300)
        plt.clf()

        plt.plot(x, rho * u, label='pysph (' + str(self.options.scheme) + ')')
        plt.plot(x_e, rho_e * u_e, label='exact')
        plt.xlabel('x')
        plt.ylabel('M')
        plt.legend()
        fig = os.path.join(self.output_dir, "Machno.png")
        plt.savefig(fig, dpi=300)
        plt.clf()

        plt.plot(x, p, label='pysph (' + str(self.options.scheme) + ')')
        plt.plot(x_e, p_e, label='exact')
        plt.xlabel('x')
        plt.ylabel('p')
        plt.legend()
        fig = os.path.join(self.output_dir, "pressure.png")
        plt.savefig(fig, dpi=300)
        plt.clf()

        fname = os.path.join(self.output_dir, 'results.npz')
        numpy.savez(fname, x=x, u=u, e=e, cs=cs, rho=rho, p=p, h=h)

        fname = os.path.join(self.output_dir, 'exact.npz')
        numpy.savez(fname, x=x_e, u=u_e, e=e_e, p=p_e, rho=rho_e)

    def configure_scheme(self):
        s = self.scheme
        dxl = 0.5/self.nl
        ratio = self.rhor/self.rhol
        nr = ratio*self.nl
        dxr = 0.5/self.nr
        h0 = self.hdx * self.dxr
        kernel_factor = self.options.hdx
        if self.options.scheme == 'mpm':
            s.configure(kernel_factor=kernel_factor)
            s.configure_solver(dt=self.dt, tf=self.tf,
                               adaptive_timestep=True, pfreq=50)
        elif self.options.scheme == 'adke':
            s.configure_solver(dt=self.dt, tf=self.tf,
                               adaptive_timestep=False, pfreq=50)
        elif self.options.scheme == 'gsph':
            s.configure_solver(dt=self.dt, tf=self.tf,
                               adaptive_timestep=False, pfreq=50)
        elif self.options.scheme == 'crk':
            s.configure_solver(dt=self.dt, tf=self.tf,
                               adaptive_timestep=False, pfreq=1)
        elif self.options.scheme in ['tsph', 'psph']:
            s.configure(hfact=kernel_factor)
            s.configure_solver(dt=self.dt, tf=self.tf,
                               adaptive_timestep=False, pfreq=50)



"""Unittests for Octree

You can run the tests like so:

    $ pytest -v run_tests.py
"""
import numpy as np

# PySPH imports
from pysph.base.utils import get_particle_array
from pysph.base.octree import Octree, CompressedOctree

# Python testing framework
import unittest
from pytest import importorskip, mark

from pysph.base.nnps import get_number_of_threads


def test_single_level_octree():
    N = 50
    x, y, z = np.mgrid[0:1:N*1j, 0:1:N*1j, 0:1:N*1j]
    x = x.ravel()
    y = y.ravel()
    z = z.ravel()
    h = np.ones_like(x)

    pa = get_particle_array(x=x, y=y, z=z, h=h)

    # For maximum leaf particles greater that total number
    # of particles
    tree = Octree(pa.get_number_of_particles() + 1)
    tree.build_tree(pa)
    # Test that depth of the tree is 1
    assert tree.depth == 1


def test_compressed_octree_has_lesser_depth_than_octree():
    N = 50
    x, y, z = np.mgrid[0:1:N*1j, 0:1:N*1j, 0:1:N*1j]
    x = x.ravel()
    y = y.ravel()
    z = z.ravel()
    h = np.ones_like(x)

    pa = get_particle_array(x=x, y=y, z=z, h=h)

    x1 = np.array([20])
    y1 = np.array([20])
    z1 = np.array([20])

    # For a dataset where one particle is far away from a
    # cluster of particles
    pa.add_particles(x=x1, y=y1, z=z1)
    tree = Octree(10)
    comp_tree = CompressedOctree(10)

    depth_tree = tree.build_tree(pa)
    depth_comp_tree = comp_tree.build_tree(pa)

    # Test that the depth of compressed octree for the same
    # leaf_max_particles is lesser than that of octree
    assert depth_comp_tree < depth_tree


def test_single_level_compressed_octree():
    N = 50
    x, y, z = np.mgrid[0:1:N*1j, 0:1:N*1j, 0:1:N*1j]
    x = x.ravel()
    y = y.ravel()
    z = z.ravel()
    h = np.ones_like(x)

    pa = get_particle_array(x=x, y=y, z=z, h=h)

    # For maximum leaf particles greater that total number
    # of particles
    tree = CompressedOctree(pa.get_number_of_particles() + 1)
    tree.build_tree(pa)
    # Test that depth of the tree is 1
    assert tree.depth == 1


class SimpleOctreeTestCase(unittest.TestCase):
    """Simple test case for Octree
    """
    def setUp(self):
        N = 50
        x, y, z = np.mgrid[0:1:N*1j, 0:1:N*1j, 0:1:N*1j]
        self.x = x.ravel()
        self.y = y.ravel()
        self.z = z.ravel()
        self.h = np.ones_like(self.x)

        self.tree = Octree(10)

    def test_levels_in_tree(self):
        pa = get_particle_array(x=self.x, y=self.y, z=self.z, h=self.h)
        self.tree.build_tree(pa)
        root = self.tree.get_root()

        def _check_levels(node, level):
            # Test that levels for nodes are correctly set
            # starting from 0 at root
            self.assertTrue(node.level == level)
            children = node.get_children()
            for child in children:
                if child is None:
                    continue
                _check_levels(child, level + 1)

        _check_levels(root, 0)
        self.tree.delete_tree()

    def test_parent_for_node(self):
        pa = get_particle_array(x=self.x, y=self.y, z=self.z, h=self.h)
        self.tree.build_tree(pa)
        root = self.tree.get_root()
        # Test that parent of root is 'None'
        self.assertTrue(root.get_parent() is None)

        def _check_parent(node):
            children = node.get_children()
            for child in children:
                if child is None:
                    continue
                # Test that the parent is set correctly for all nodes
                self.assertTrue(child.get_parent() == node)
                _check_parent(child)

        _check_parent(root)
        self.tree.delete_tree()

    def test_sum_of_indices_lengths_equals_total_number_of_particles(self):
        pa = get_particle_array(x=self.x, y=self.y, z=self.z, h=self.h)
        self.tree.build_tree(pa)
        root = self.tree.get_root()
        sum_indices = [0]

        def _calculate_sum(node, sum_indices):
            indices = node.get_indices(self.tree)
            sum_indices[0] += indices.length
            children = node.get_children()
            for child in children:
                if child is None:
                    continue
                _calculate_sum(child, sum_indices)

        _calculate_sum(root, sum_indices)
        # Test that sum of lengths of all indices is equal to total
        # number of particles
        self.assertTrue(pa.get_number_of_particles() == sum_indices[0])
        self.tree.delete_tree()

    def test_all_indices_are_unique(self):
        pa = get_particle_array(x=self.x, y=self.y, z=self.z, h=self.h)
        self.tree.build_tree(pa)
        root = self.tree.get_root()
        all_indices = []

        def _find_indices(node):
            indices = node.get_indices(self.tree)
            for i in indices:
                all_indices.append(i)
            children = node.get_children()
            for child in children:
                if child is None:
                    continue
                _find_indices(child)

        _find_indices(root)
        # Test that indices are unique
        self.assertTrue(len(all_indices) == len(set(all_indices)))
        self.tree.delete_tree()

    def test_plot_root(self):
        pa = get_particle_array(x=self.x, y=self.y, z=self.z, h=self.h)
        self.tree.build_tree(pa)
        root = self.tree.get_root()
        mpl = importorskip("matplotlib")
        mpl.use('Agg')
        import matplotlib.pyplot as plt
        from mpl_toolkits.mplot3d import Axes3D
        fig = plt.figure()
        ax = Axes3D(fig)

        root.plot(ax)

        for line in ax.lines:
            xs = line.get_xdata()
            ys = line.get_ydata()
            line_length = (xs[0] - xs[1])**2 + (ys[0] - ys[1])**2
            # Test that the lengths of sides 2D projection of the node
            # is equal to the length of the side of the node
            self.assertTrue(line_length == root.length**2 or
                            line_length == 0)

        self.tree.delete_tree()


@mark.skipif(get_number_of_threads() == 1,
             reason="N_threads=1; OpenMP does not seem available.")
def test_single_level_parallel_octree():
    N = 50
    x, y, z = np.mgrid[0:1:N*1j, 0:1:N*1j, 0:1:N*1j]
    x = x.ravel()
    y = y.ravel()
    z = z.ravel()
    h = np.ones_like(x)

    pa = get_particle_array(x=x, y=y, z=z, h=h)

    # For maximum leaf particles greater that total number
    # of particles
    tree = Octree(pa.get_number_of_particles() + 1)
    tree.build_tree(pa, True)
    # Test that depth of the tree is 1
    assert tree.depth == 1


@mark.skipif(get_number_of_threads() == 1,
             reason="N_threads=1; OpenMP does not seem available.")
def test_single_level_parallel_compressed_octree():
    N = 50
    x, y, z = np.mgrid[0:1:N*1j, 0:1:N*1j, 0:1:N*1j]
    x = x.ravel()
    y = y.ravel()
    z = z.ravel()
    h = np.ones_like(x)

    pa = get_particle_array(x=x, y=y, z=z, h=h)

    # For maximum leaf particles greater that total number
    # of particles
    tree = CompressedOctree(pa.get_number_of_particles() + 1)
    tree.build_tree(pa, True)
    # Test that depth of the tree is 1
    assert tree.depth == 1


@mark.skipif(get_number_of_threads() == 1,
             reason="N_threads=1; OpenMP does not seem available.")
def test_parallel_compressed_octree_has_lesser_depth_than_parallel_octree():
    N = 50
    x, y, z = np.mgrid[0:1:N*1j, 0:1:N*1j, 0:1:N*1j]
    x = x.ravel()
    y = y.ravel()
    z = z.ravel()
    h = np.ones_like(x)

    pa = get_particle_array(x=x, y=y, z=z, h=h)

    x1 = np.array([20])
    y1 = np.array([20])
    z1 = np.array([20])

    # For a dataset where one particle is far away from a
    # cluster of particles
    pa.add_particles(x=x1, y=y1, z=z1)
    tree = Octree(10)
    comp_tree = CompressedOctree(10)

    depth_tree = tree.build_tree(pa, True)
    depth_comp_tree = comp_tree.build_tree(pa, True)

    # Test that the depth of compressed octree for the same
    # leaf_max_particles is lesser than that of octree
    assert depth_comp_tree < depth_tree


@mark.skipif(get_number_of_threads() == 1,
             reason="N_threads=1; OpenMP does not seem available.")
class SimpleParallelOctreeTestCase(unittest.TestCase):
    """Simple test case for Parallel Octree
    """
    def setUp(self):
        N = 50
        x, y, z = np.mgrid[0:1:N*1j, 0:1:N*1j, 0:1:N*1j]
        self.x = x.ravel()
        self.y = y.ravel()
        self.z = z.ravel()
        self.h = np.ones_like(self.x)

        self.tree = Octree(10)

    def test_levels_in_tree(self):
        pa = get_particle_array(x=self.x, y=self.y, z=self.z, h=self.h)
        self.tree.build_tree(pa, True)
        root = self.tree.get_root()

        def _check_levels(node, level):
            # Test that levels for nodes are correctly set
            # starting from 0 at root
            self.assertTrue(node.level == level)
            children = node.get_children()
            for child in children:
                if child is None:
                    continue
                _check_levels(child, level + 1)

        _check_levels(root, 0)
        self.tree.delete_tree()

    def test_parent_for_node(self):
        pa = get_particle_array(x=self.x, y=self.y, z=self.z, h=self.h)
        self.tree.build_tree(pa, True)
        root = self.tree.get_root()
        # Test that parent of root is 'None'
        self.assertTrue(root.get_parent() is None)

        def _check_parent(node):
            children = node.get_children()
            for child in children:
                if child is None:
                    continue
                # Test that the parent is set correctly for all nodes
                self.assertTrue(child.get_parent() == node)
                _check_parent(child)

        _check_parent(root)
        self.tree.delete_tree()

    def test_sum_of_indices_lengths_equals_total_number_of_particles(self):
        pa = get_particle_array(x=self.x, y=self.y, z=self.z, h=self.h)
        self.tree.build_tree(pa, True)
        root = self.tree.get_root()
        sum_indices = [0]

        def _calculate_sum(node, sum_indices):
            indices = node.get_indices(self.tree)
            sum_indices[0] += indices.length
            children = node.get_children()
            for child in children:
                if child is None:
                    continue
                _calculate_sum(child, sum_indices)

        _calculate_sum(root, sum_indices)
        # Test that sum of lengths of all indices is equal to total
        # number of particles
        self.assertTrue(pa.get_number_of_particles() == sum_indices[0])
        self.tree.delete_tree()

    def test_all_indices_are_unique(self):
        pa = get_particle_array(x=self.x, y=self.y, z=self.z, h=self.h)
        self.tree.build_tree(pa, True)
        root = self.tree.get_root()
        all_indices = []

        def _find_indices(node):
            indices = node.get_indices(self.tree)
            for i in indices:
                all_indices.append(i)
            children = node.get_children()
            for child in children:
                if child is None:
                    continue
                _find_indices(child)

        _find_indices(root)
        # Test that indices are unique
        self.assertTrue(len(all_indices) == len(set(all_indices)))
        self.tree.delete_tree()

    def test_plot_root(self):
        pa = get_particle_array(x=self.x, y=self.y, z=self.z, h=self.h)
        self.tree.build_tree(pa, True)
        root = self.tree.get_root()
        mpl = importorskip("matplotlib")
        mpl.use('Agg')
        import matplotlib.pyplot as plt
        from mpl_toolkits.mplot3d import Axes3D
        fig = plt.figure()
        ax = Axes3D(fig)

        root.plot(ax)

        for line in ax.lines:
            xs = line.get_xdata()
            ys = line.get_ydata()
            line_length = (xs[0] - xs[1])**2 + (ys[0] - ys[1])**2
            # Test that the lengths of sides 2D projection of the node
            # is equal to the length of the side of the node
            self.assertTrue(line_length == root.length**2 or
                            line_length == 0)

        self.tree.delete_tree()


class TestOctreeFor2DDataset(SimpleOctreeTestCase):
    """Test Octree for 2D dataset
    """
    def setUp(self):
        N = 500
        x, y = np.mgrid[0:1:N*1j, 0:1:N*1j]
        self.x = x.ravel()
        self.y = y.ravel()
        self.z = np.zeros_like(self.x)
        self.h = np.ones_like(self.x)

        self.tree = Octree(10)


class TestOctreeFor1DDataset(SimpleOctreeTestCase):
    """Test Octree for 1D dataset
    """
    def setUp(self):
        N = int(1e5)
        self.x = np.linspace(0, 1, num=N)
        self.y = np.zeros_like(self.x)
        self.z = np.zeros_like(self.x)
        self.h = np.ones_like(self.x)

        self.tree = Octree(100)


class TestOctreeForFloatingPointError(SimpleOctreeTestCase):
    """Test Octree for floating point error
    """
    def setUp(self):
        N = 50
        x, y, z = np.mgrid[-1:1:N*1j, -1:1:N*1j, -1:1:N*1j]
        self.x = x.ravel()
        self.y = y.ravel()
        self.z = z.ravel()

        x1 = np.array([-1e-20])
        y1 = np.array([1e-20])
        z1 = np.array([1e-20])

        self.x = np.concatenate([self.x, x1])
        self.y = np.concatenate([self.y, y1])
        self.z = np.concatenate([self.z, z1])

        self.h = np.ones_like(self.x)
        self.tree = Octree(10)


@mark.skipif(get_number_of_threads() == 1,
             reason="N_threads=1; OpenMP does not seem available.")
class TestParallelOctreeFor2DDataset(SimpleParallelOctreeTestCase):
    """Test Octree for 2D dataset
    """
    def setUp(self):
        N = 500
        x, y = np.mgrid[0:1:N*1j, 0:1:N*1j]
        self.x = x.ravel()
        self.y = y.ravel()
        self.z = np.zeros_like(self.x)
        self.h = np.ones_like(self.x)

        self.tree = Octree(10)


@mark.skipif(get_number_of_threads() == 1,
             reason="N_threads=1; OpenMP does not seem available.")
class TestParallelOctreeFor1DDataset(SimpleParallelOctreeTestCase):
    """Test Octree for 1D dataset
    """
    def setUp(self):
        N = int(1e5)
        self.x = np.linspace(0, 1, num=N)
        self.y = np.zeros_like(self.x)
        self.z = np.zeros_like(self.x)
        self.h = np.ones_like(self.x)

        self.tree = Octree(100)


@mark.skipif(get_number_of_threads() == 1,
             reason="N_threads=1; OpenMP does not seem available.")
class TestParallelOctreeForFloatingPointError(SimpleParallelOctreeTestCase):
    """Test Octree for floating point error
    """
    def setUp(self):
        N = 50
        x, y, z = np.mgrid[-1:1:N*1j, -1:1:N*1j, -1:1:N*1j]
        self.x = x.ravel()
        self.y = y.ravel()
        self.z = z.ravel()

        x1 = np.array([-1e-20])
        y1 = np.array([1e-20])
        z1 = np.array([1e-20])

        self.x = np.concatenate([self.x, x1])
        self.y = np.concatenate([self.y, y1])
        self.z = np.concatenate([self.z, z1])

        self.h = np.ones_like(self.x)
        self.tree = Octree(10)


class SimpleCompressedOctreeTestCase(SimpleOctreeTestCase):
    """Simple test case for Compressed Octree
    """
    def setUp(self):
        N = 50
        x, y, z = np.mgrid[0:1:N*1j, 0:1:N*1j, 0:1:N*1j]
        self.x = x.ravel()
        self.y = y.ravel()
        self.z = z.ravel()
        self.h = np.ones_like(self.x)

        self.tree = CompressedOctree(10)


class TestCompressedOctreeFor1DDataset(SimpleOctreeTestCase):
    """Test Octree for 1D dataset
    """
    def setUp(self):
        N = int(1e5)
        self.x = np.linspace(0, 1, num=N)
        self.y = np.zeros_like(self.x)
        self.z = np.zeros_like(self.x)
        self.h = np.ones_like(self.x)

        self.tree = CompressedOctree(100)


class TestCompressedOctreeFor2DDataset(SimpleOctreeTestCase):
    """Test Octree for 2D dataset
    """
    def setUp(self):
        N = 500
        x, y = np.mgrid[0:1:N*1j, 0:1:N*1j]
        self.x = x.ravel()
        self.y = y.ravel()
        self.z = np.zeros_like(self.x)
        self.h = np.ones_like(self.x)

        self.tree = CompressedOctree(10)


class TestCompressedOctreeForFloatingPointError(SimpleOctreeTestCase):
    """Test Octree for floating point error
    """
    def setUp(self):
        N = 50
        x, y, z = np.mgrid[-1:1:N*1j, -1:1:N*1j, -1:1:N*1j]
        self.x = x.ravel()
        self.y = y.ravel()
        self.z = z.ravel()

        x1 = np.array([-1e-20])
        y1 = np.array([1e-20])
        z1 = np.array([1e-20])

        self.x = np.concatenate([self.x, x1])
        self.y = np.concatenate([self.y, y1])
        self.z = np.concatenate([self.z, z1])

        self.h = np.ones_like(self.x)

        self.tree = CompressedOctree(10)


@mark.skipif(get_number_of_threads() == 1,
             reason="N_threads=1; OpenMP does not seem available.")
class SimpleParallelCompressedOctreeTestCase(SimpleParallelOctreeTestCase):
    """Simple test case for Compressed Octree
    """
    def setUp(self):
        N = 50
        x, y, z = np.mgrid[0:1:N*1j, 0:1:N*1j, 0:1:N*1j]
        self.x = x.ravel()
        self.y = y.ravel()
        self.z = z.ravel()
        self.h = np.ones_like(self.x)

        self.tree = CompressedOctree(10)


@mark.skipif(get_number_of_threads() == 1,
             reason="N_threads=1; OpenMP does not seem available.")
class TestParallelCompressedOctreeFor1DDataset(SimpleParallelOctreeTestCase):
    """Test Octree for 1D dataset
    """
    def setUp(self):
        N = int(1e5)
        self.x = np.linspace(0, 1, num=N)
        self.y = np.zeros_like(self.x)
        self.z = np.zeros_like(self.x)
        self.h = np.ones_like(self.x)

        self.tree = CompressedOctree(100)


@mark.skipif(get_number_of_threads() == 1,
             reason="N_threads=1; OpenMP does not seem available.")
class TestParallelCompressedOctreeFor2DDataset(SimpleParallelOctreeTestCase):
    """Test Octree for 2D dataset
    """
    def setUp(self):
        N = 500
        x, y = np.mgrid[0:1:N*1j, 0:1:N*1j]
        self.x = x.ravel()
        self.y = y.ravel()
        self.z = np.zeros_like(self.x)
        self.h = np.ones_like(self.x)

        self.tree = CompressedOctree(10)


@mark.skipif(get_number_of_threads() == 1,
             reason="N_threads=1; OpenMP does not seem available.")
class TestParallelCompressedOctreeForFloatingPointError(
        SimpleParallelOctreeTestCase):
    """Test Octree for floating point error
    """
    def setUp(self):
        N = 50
        x, y, z = np.mgrid[-1:1:N*1j, -1:1:N*1j, -1:1:N*1j]
        self.x = x.ravel()
        self.y = y.ravel()
        self.z = z.ravel()

        x1 = np.array([-1e-20])
        y1 = np.array([1e-20])
        z1 = np.array([1e-20])

        self.x = np.concatenate([self.x, x1])
        self.y = np.concatenate([self.y, y1])
        self.z = np.concatenate([self.z, z1])

        self.h = np.ones_like(self.x)

        self.tree = CompressedOctree(10)


if __name__ == '__main__':
    unittest.main()



"""Lid driven cavity using the Transport Velocity formulation. (10 minutes)
"""

import os

# PySPH imports
from pysph.base.utils import get_particle_array
from pysph.solver.application import Application

from pysph.sph.scheme import TVFScheme, SchemeChooser
from pysph.sph.wc.edac import EDACScheme

# numpy
import numpy as np

# domain and reference values
L = 1.0
Umax = 1.0
c0 = 10 * Umax
rho0 = 1.0
p0 = c0 * c0 * rho0

# Numerical setup
hdx = 1.0


class LidDrivenCavity(Application):
    def add_user_options(self, group):
        group.add_argument(
            "--nx", action="store", type=int, dest="nx", default=50,
            help="Number of points along x direction."
        )
        group.add_argument(
            "--re", action="store", type=float, dest="re", default=100,
            help="Reynolds number (defaults to 100)."
        )
        self.n_avg = 5
        group.add_argument(
            "--n-vel-avg", action="store", type=int, dest="n_avg",
            default=None,
            help="Average velocities over these many saved timesteps."
        )

    def consume_user_options(self):
        nx = self.options.nx
        if self.options.n_avg is not None:
            self.n_avg = self.options.n_avg
        self.dx = L / nx
        self.re = self.options.re
        h0 = hdx * self.dx
        self.nu = Umax * L / self.re
        dt_cfl = 0.25 * h0 / (c0 + Umax)
        dt_viscous = 0.125 * h0**2 / self.nu
        dt_force = 1.0
        self.tf = 10.0
        self.dt = min(dt_cfl, dt_viscous, dt_force)

    def configure_scheme(self):
        h0 = hdx * self.dx
        if self.options.scheme == 'tvf':
            self.scheme.configure(h0=h0, nu=self.nu)
        elif self.options.scheme == 'edac':
            self.scheme.configure(h=h0, nu=self.nu)
        self.scheme.configure_solver(tf=self.tf, dt=self.dt, pfreq=500)

    def create_scheme(self):
        tvf = TVFScheme(
            ['fluid'], ['solid'], dim=2, rho0=rho0, c0=c0, nu=None,
            p0=p0, pb=p0, h0=hdx
        )
        edac = EDACScheme(
            fluids=['fluid'], solids=['solid'], dim=2, c0=c0, rho0=rho0,
            nu=0.0, pb=p0, eps=0.0, h=0.0,
        )
        s = SchemeChooser(default='tvf', tvf=tvf, edac=edac)
        return s

    def create_particles(self):
        dx = self.dx
        ghost_extent = 5 * dx
        # create all the particles
        _x = np.arange(-ghost_extent - dx / 2, L + ghost_extent + dx / 2, dx)
        x, y = np.meshgrid(_x, _x)
        x = x.ravel()
        y = y.ravel()

        # sort out the fluid and the solid
        indices = []
        for i in range(x.size):
            if ((x[i] > 0.0) and (x[i] < L)):
                if ((y[i] > 0.0) and (y[i] < L)):
                    indices.append(i)

        # create the arrays
        solid = get_particle_array(name='solid', x=x, y=y)

        # remove the fluid particles from the solid
        fluid = solid.extract_particles(indices)
        fluid.set_name('fluid')
        solid.remove_particles(indices)

        print("Lid driven cavity :: Re = %d, dt = %g" % (self.re, self.dt))

        # setup the particle properties
        volume = dx * dx

        # mass is set to get the reference density of rho0
        fluid.m[:] = volume * rho0
        solid.m[:] = volume * rho0
        # Set a reference rho also, some schemes will overwrite this with a
        # summation density.
        fluid.rho[:] = rho0
        solid.rho[:] = rho0

        # smoothing lengths
        fluid.h[:] = hdx * dx
        solid.h[:] = hdx * dx

        # imposed horizontal velocity on the lid
        solid.u[:] = 0.0
        solid.v[:] = 0.0

        for i in range(solid.get_number_of_particles()):
            if solid.y[i] > L:
                solid.u[i] = Umax

        # add requisite properties to the arrays:
        self.scheme.setup_properties([fluid, solid])

        # volume is set as dx^2
        fluid.V[:] = 1. / volume
        solid.V[:] = 1. / volume

        return [fluid, solid]

    def post_process(self, info_fname):
        try:
            import matplotlib
            matplotlib.use('Agg')
            from matplotlib import pyplot as plt
        except ImportError:
            print("Post processing requires matplotlib.")
            return
        if self.rank > 0:
            return
        info = self.read_info(info_fname)
        if len(self.output_files) == 0:
            return
        t, ke = self._plot_ke_history()
        x, ui, vi, ui_c, vi_c = self._plot_velocity()
        res = os.path.join(self.output_dir, "results.npz")
        np.savez(res, t=t, ke=ke, x=x, u=ui, v=vi, u_c=ui_c, v_c=vi_c)

    def _plot_ke_history(self):
        from pysph.tools.pprocess import get_ke_history
        from matplotlib import pyplot as plt
        t, ke = get_ke_history(self.output_files, 'fluid')
        plt.clf()
        plt.plot(t, ke)
        plt.xlabel('t')
        plt.ylabel('Kinetic energy')
        fig = os.path.join(self.output_dir, "ke_history.png")
        plt.savefig(fig, dpi=300)
        return t, ke

    def _plot_velocity(self):
        from pysph.tools.interpolator import Interpolator
        from pysph.solver.utils import load
        from pysph.examples.ghia_cavity_data import get_u_vs_y, get_v_vs_x
        # interpolated velocities
        _x = np.linspace(0, 1, 101)
        xx, yy = np.meshgrid(_x, _x)

        # take the last solution data
        fname = self.output_files[-1]
        data = load(fname)
        tf = data['solver_data']['t']
        interp = Interpolator(list(data['arrays'].values()), x=xx, y=yy)
        ui = np.zeros_like(xx)
        vi = np.zeros_like(xx)
        # Average out the velocities over the last n_avg timesteps
        for fname in self.output_files[-self.n_avg:]:
            data = load(fname)
            tf = data['solver_data']['t']
            interp.update_particle_arrays(list(data['arrays'].values()))
            _u = interp.interpolate('u')
            _v = interp.interpolate('v')
            _u.shape = 101, 101
            _v.shape = 101, 101
            ui += _u
            vi += _v

        ui /= self.n_avg
        vi /= self.n_avg

        # velocity magnitude
        self.vmag = vmag = np.sqrt(ui**2 + vi**2)
        import matplotlib.pyplot as plt

        f = plt.figure()

        plt.streamplot(
            xx, yy, ui, vi, density=(2, 2),  # linewidth=5*vmag/vmag.max(),
            color=vmag
        )
        plt.xlim(0, 1)
        plt.ylim(0, 1)
        plt.colorbar()
        plt.xlabel('$x$')
        plt.ylabel('$y$')
        plt.title('Streamlines at %s seconds' % tf)
        fig = os.path.join(self.output_dir, 'streamplot.png')
        plt.savefig(fig, dpi=300)

        f = plt.figure()

        ui_c = ui[:, 50]
        vi_c = vi[50]

        s1 = plt.subplot(211)
        s1.plot(ui_c, _x, label='Computed')

        y, data = get_u_vs_y()
        if self.re in data:
            s1.plot(data[self.re], y, 'o', fillstyle='none',
                    label='Ghia et al.')
        s1.set_xlabel(r'$v_x$')
        s1.set_ylabel(r'$y$')
        s1.legend()

        s2 = plt.subplot(212)
        s2.plot(_x, vi_c, label='Computed')
        x, data = get_v_vs_x()
        if self.re in data:
            s2.plot(x, data[self.re], 'o', fillstyle='none',
                    label='Ghia et al.')
        s2.set_xlabel(r'$x$')
        s2.set_ylabel(r'$v_y$')
        s2.legend()

        fig = os.path.join(self.output_dir, 'centerline.png')
        plt.savefig(fig, dpi=300)
        return _x, ui, vi, ui_c, vi_c


if __name__ == '__main__':
    app = LidDrivenCavity()
    app.run()
    app.post_process(app.info_filename)
